/*
 *  cold reset from 0xFFFF:0 (last 16 bytes of rom)
 *  fixes UMCS for 256 kB rom and
 *  jumps to here, boot: in 0xC000:0x0100
 */

#include "head.h"
#include "pcb.h"
#include "asic.h"

#define _CS 0       /* both in the 128 kB ram */
#define _DS 0x1000

#asm
	include asm.inc

	export _version
	export _timex   ; seconds, in different formats
	export _timew
	export _timef
	export _time    ; dword time()
	export _fido

;----------------------------------------------------------------------
	.data
__data:
	.long 0xABADACCE

	.long __text
	.long __data
	.long __edata
_brk:
	.long __end

_romseg:
	.word 0xC000    ; 8000, C000 or E000

_version:
	.asciz VERSION

task:
	.word task_0

task_free:
	.word 0

tsk_next equ 0
tsk_sp   equ 2
tsk_ss   equ 4
tsk_zzz  equ 6      ; unused
tsk_size equ 8

task_0:
	.word task_0    ; next.   now only task_0 exists
	.word 0, _DS    ; sp, ss. filled when switching out
	.word 0         ; unused

;----------------------------------------------------------------------
	.bss

	.comm junk, 2   ; safe location for DMA pointers

_timex:             ; 32.16
	.space 1
_timef:             ; 8.8
	.space 1
_timew:             ; 16.0
	.space 4        ; _time() for 32.0

_fido:
	.space 2        ; watchdog downcount

;----------------------------------------------------------------------
	.text
__text:

	entry boot
	export _reboot
	import _initialise
	import _mainloop
	import _tick

;----------------------------------------------------------------------
; 256 interrupt vectors at 0:0
; all handlers currently in _CS segment (valid after first movsb, below)

	macro ivec
	.word ?1, _CS
	mend

	ivec divide_error      ; 00 0000
	ivec single_step       ; 01 0004
v_nmi:
	ivec nmi               ; 02 0008
	ivec breakpoint        ; 03 000C
	ivec overflow          ; 04 0010
	ivec bounds_check      ; 05 0014
	ivec invalid_opcode    ; 06 0018
	ivec floating_escape   ; 07 001C
	ivec timer_0_int       ; 08 0020
	ivec reserved          ; 09 0024
	ivec dma_0_int         ; 0A 0028
	ivec dma_1_int         ; 0B 002C
	ivec int_0             ; 0C 0030
	ivec int_1             ; 0D 0034
	ivec int_2             ; 0E 0038
	ivec int_3             ; 0F 003C
	ivec reserved          ; 10 0040 (Numerics)
	ivec reserved          ; 11 0044
	ivec timer_1_int       ; 12 0048
	ivec timer_2_int       ; 13 004C
	ivec reserved          ; 14 0050
	ivec reserved          ; 15 0054
	ivec reserved          ; 16 0058
	ivec reserved          ; 17 005C
	ivec reserved          ; 18 0060
	ivec reserved          ; 19 0064
	ivec reserved          ; 1A 0068
	ivec reserved          ; 1B 006C
	ivec reserved          ; 1C 0070
	ivec reserved          ; 1D 0074
	ivec reserved          ; 1E 0078
	ivec reserved          ; 1F 007C last reserved

	ivec exit              ; 20 0080 available for user
	ivec unknown           ; 21 0084
	ivec new_task          ; 22 0088
	ivec yield             ; 23 008C
	ivec unknown           ; 24 0090
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown           ; 2F 00BC

	macro eight_unknown_ivecs
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	mend

	eight_unknown_ivecs    ; 30 00C0
	eight_unknown_ivecs    ; 38 00E0
	eight_unknown_ivecs    ; 40 0100
	eight_unknown_ivecs    ; 48 0120
	eight_unknown_ivecs    ; 50 0140
	eight_unknown_ivecs    ; 58 0160
	eight_unknown_ivecs    ; 60 0180
	eight_unknown_ivecs    ; 68 01A0
	eight_unknown_ivecs    ; 70 01C0
	eight_unknown_ivecs    ; 78 01E0
	eight_unknown_ivecs    ; 80 0200
	eight_unknown_ivecs    ; 88 0220
	eight_unknown_ivecs    ; 90 0240
	eight_unknown_ivecs    ; 98 0260
	eight_unknown_ivecs    ; A0 0280
	eight_unknown_ivecs    ; A8 02A0
	eight_unknown_ivecs    ; B0 02C0
	eight_unknown_ivecs    ; B8 02E0
	eight_unknown_ivecs    ; C0 0300
	eight_unknown_ivecs    ; C8 0320
	eight_unknown_ivecs    ; D0 0340
	eight_unknown_ivecs    ; D8 0360
	eight_unknown_ivecs    ; E0 0380
	eight_unknown_ivecs    ; E8 03A0
	eight_unknown_ivecs    ; F0 03C0
	eight_unknown_ivecs    ; F8 03E0

	;-------------------------------------------------
	; reset fixes SS:SP to 0:0,
	; UCS to 256 kB, C0000 ... FFFFF. R=101: 1 waitstate, no RDY,
	; and jumps to C000:0400 (boot)
	;
	; things should be straightened out in less than
	; one second, watchdog nmi.

boot:                  ; absolute address in ROM C000:0400
retry:                 ; relative jumps from below
	cli
	cld

	;-------------------------------------------------
	; memory and io chipselects

	; PCS0 at 0000, R=111: 3 waitstates, no RDY, for PCS0-3

	mov ax, # 0x003F
	mov dx, # PACS
	out dx, PCB_AX

	; MCS0 at 00000, R=001: 1 waitstate, external RDY, for MCS0-3

	mov ax, # 0x01F9        ; XXX external RDY must be used
	mov dx, # MMCS
	out dx, PCB_AX

	; MCS0-3: four times 128 kB = 512 kB
	; EX,MS=10: all PCS0-6 in i/o space
	; R=010: 2 waitstates, external RDY for PCS4-6 XXX

	mov ax, # 0xC0BA
	mov dx, # MPCS
	out dx, PCB_AX

	; MCS0 is at zero, so LCS is not/can not be used.
	; LMCS is not programmed

	;-------------------------------------------------
	; relocate from  C000:0 to ram,
	; text              0:0 64 kB and
	; data/bss/stack 1000:0 64 kB

	mov dx, # _DS      ; final SS:SP  _DS
	mov ss, dx
	mov sp, # 0        ; wraps to ss:0xFFFE on first push

	mov ax, cs         ; ROM: 0x8000, 0xC000 or 0xE000
	mov ds, ax
	xor si, si         ; src 0xC000:0

	mov ax, # _CS
	mov es, ax
	xor di, di         ; dst 0:0

	mov bx, # __etext
	mov cx, bx
	rep
	movsb              ; copy text. interrupt vectors (nmi) now valid.

	xor si, si         ; rewind
	xor di, di

	mov cx, bx         ; __etext
	rep
	cmpsb              ; verify text
	jnz retry          ; while bad copy (Vcc coming up?)

	; C000 + 1000 + F029 (if 29 paragraphs of text)

	mov ax, cs         ; ROM
	add ax, dx         ; _DS
	add ax, # __segoff ; includes -_DS
	mov ds, ax
	xor si, si         ; src (0xC000 + segoff):0

	mov es, dx         ; _DS
	xor di, di         ; dst 0x1000:0

	mov bx, # __edata
	mov cx, bx
	rep
	movsb              ; copy data XXX remember end in rom, ds:si

	xor si, si         ; rewind
	xor di, di

	mov cx, bx         ; __edata
	rep
	cmpsb              ; verify data
	jnz retry          ; while bad copy

	xor ax, ax

	mov cx, di         ; end of initialized data
	neg cx             ; to end of segment
	rep
	stosb              ; zero bss and stack

	mov di, bx         ; rewind, to __edata

	mov cx, di
	neg cx
	rep
	scasb              ; verify zero
	jnz retry          ; while bad copy

	mov ds, dx         ; final DS  _DS
	mov es, dx         ; final ES  _DS

	mov _romseg, cs    ; ROM is all but forgotten from now on.

	jmpf .2, _CS       ; final CS  _CS
.2:
	;-------------------------------------------------
	; interrupt controller

	mov ax, # 0xFFFF   ; opened by each MSK=0 below
	mov dx, # IMASK
	out dx, PCB_AX

	mov ax, # 7        ; no levels masked
	mov dx, # IPRIO
	out dx, PCB_AX

	mov cx, # 256      ; flush pending interrupts XXX meaningless
	mov ax, # EOI_ANY
	mov dx, # EOI
.3:
	out dx, PCB_AX
	loop .3

	;-------------------------------------------------
	; external interrupts
	; uarts etc from asic
	; XXX edge vs level ?

	mov ax, # 0x10    ; level-trigger MSK=0 PR=0
	mov dx, # I0CON
	out dx, PCB_AX

	mov ax, # 0x13    ; level-trigger MSK=0 PR=3
	mov dx, # I1CON
	out dx, PCB_AX

	mov ax, # 0x11    ; level-trigger MSK=0 PR=1
	mov dx, # I2CON
	out dx, PCB_AX

	mov ax, # 0x14    ; level-trigger MSK=0 PR=4
	mov dx, # I3CON
	out dx, PCB_AX

	;-------------------------------------------------
	; Timer 1
	; freerunning, about 2.032 MHz, period about 32 msec

	mov ax, # 0xC001  ; EN /INH CONT
	mov dx, # T1MODE
	out dx, PCB_AX

	mov ax, # 0       ; 0x10000
	mov dx, # T1CMP
	out dx, PCB_AX

	;-------------------------------------------------
	; Timer 2
	; periodic interval timer, 256 Hz

	mov ax, # 0x6001  ; /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	mov ax, # (TMRCLK + HZ/2) / HZ    ; should be close enough
	mov dx, # T2CMP
	out dx, PCB_AX

	mov ax, # 0
	mov dx, # T2CNT
	out dx, PCB_AX

	mov ax, # 0xE001  ; EN /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	;-------------------------------------------------
	; Timers 0, 1 and 2

	mov ax, # 2       ; MSK=0 PR=2
	mov dx, # TICON
	out dx, PCB_AX

	;-------------------------------------------------
	; DMA controller, low priority tick

	mov bx, # _DS >> 12
	mov ax, # junk
	mov dx, # D0SRCL
	out dx, PCB_AX
	mov dx, # D0DSTL
	out dx, PCB_AX
	mov dx, # D1SRCL
	out dx, PCB_AX
	mov dx, # D1DSTL
	out dx, PCB_AX
	mov ax, bx
	mov dx, # D0SRCH
	out dx, PCB_AX
	mov dx, # D0DSTH
	out dx, PCB_AX
	mov dx, # D1SRCH
	out dx, PCB_AX
	mov dx, # D1DSTH
	out dx, PCB_AX         ; &junk in all pointers

	mov ax, # 1
	mov dx, # D0CNT
	out dx, PCB_AX
	mov dx, # D1CNT
	out dx, PCB_AX

	; memory to memory, constant pointers, TC INT, byte xfer

	mov ax, # 0x9370      ; Timer 2, P
	mov dx, # D0CON
	out dx, PCB_AX

	mov ax, # 0x9300      ; Unsynchronized
	mov dx, # D1CON
	out dx, PCB_AX

	mov ax, # 8 | 0       ; MSK=1 PR=0
	mov dx, # D0ICON
	out dx, PCB_AX

	mov ax, # 0 | 7       ; MSK=0 PR=7
	mov dx, # D1ICON
	out dx, PCB_AX

	;-------------------------------------------------

	xor bp, bp            ; marks end of chain

	call _initialise
	sti

main:
	mov _fido, # HZ * 10  ; 10 seconds

	call _mainloop
	jmp main

;----------------------------------------------------------------------

	jmpf 0, 0xFFFF    ; the cold start. not used

;----------------------------------------------------------------------
; reasons:
;  power off, time left about 35 msec (if large backlight off)
;  1 second since last watchdog rewind, time left to reset 1 second

	.data
nmi_msg: .asciz "nmi"

	.text
nmi:
	pusha
	in al, PA_DAT
	test al, OFFBUTTON
	jnz offia                ; #on/off, wait silently

	mov ax, # nmi_msg
	jmp near crash

offia:
	mov al, # 0
	out PA_DAT, al           ; DISP_ON off

_reboot:

	; avoid nmi cascade

	mov ax, # _DS
	mov ss, ax
	mov sp, # 0

	mov ax, # 0
	mov ds, ax
	mov [v_nmi+0], # _reboot
	mov [v_nmi+2], cs

	; force hardware reset from watchdog.
.9:
	cli
	hlt
	jmp .9

;----------------------------------------------------------------------
; unknown software interrupt

unknown:
	sti
	push bp
	mov bp, sp       ; bp ip cs flags
	or [bp+6], # 1   ; set carry
	mov ax, # -1     ; error code -1
	pop bp
	iret

;----------------------------------------------------------------------
; software interrupt 0x20
;
; no args, no return

exit:                       ; interrupts left disabled throughout
	mov ax, # _DS
	mov ds, ax

	mov si, task
	mov ax, si
.8:
	mov bx, ax
	mov ax, [bx+tsk_next]
	cmp ax, si              ; bx is previous of si ?
	jne .8

	mov ax, [si+tsk_next]
	mov [bx+tsk_next], ax   ; removed from chain

	mov ax, task_free       ; XXX task_relse:
	mov [si+tsk_next], ax
	mov task_free, si

	jmp near ret_bx         ; switch to the previous

;----------------------------------------------------------------------
; software interrupt 0x22
;
; arguments in stack:
;  ip_entry, cs, sp_empty, ss, ds_es
;
; returns:
;  dx:ax task structure
;
; bx, cx trashed

new_task:
	sti                     ; benign stuff first
	push bp
	mov bp, sp              ; bp ip cs flags ip cs sp ss dses
	push si                 ; 0  2  4  6     8  10 12 14 16
	push ds

	; prepare final return (XXX far return required)
	; and initial registers to stacktop

	mov ax, [bp+8]          ; ip_entry
	mov cx, [bp+10]         ; cs
	lds bx, [bp+12]         ; sp_empty ss
	mov dx, [bp+16]         ; ds_es

	sub bx, # 4             ; return from task
	mov [bx+0], # exit
	mov [bx+2], cs

	sub bx, # IREGS_SZ      ; initial regs

	mov [bx+IREGS_DS], dx   ; DS=ES
	mov [bx+IREGS_ES], dx

	mov [bx+IREGS_IP], ax
	mov [bx+IREGS_CS], cx
	mov [bx+IREGS_FL], # 0x200 ; IF. interrupts will be enabled

	xor ax, ax              ; lots of zeroes
	mov [bx+IREGS_AX], ax
	mov [bx+IREGS_BX], ax
	mov [bx+IREGS_CX], ax
	mov [bx+IREGS_DX], ax
	mov [bx+IREGS_SI], ax
	mov [bx+IREGS_DI], ax
	mov [bx+IREGS_BP], ax

	mov cx, ds              ; stacktop in cx:dx
	mov dx, bx

	; link to task chain

	mov ax, # _DS
	mov ds, ax

	cli

	mov bx, task_free       ; XXX task_alloc:
	or bx, bx
	jz .5                   ; must allocate new task structure
	mov ax, [bx+tsk_next]
	mov task_free, ax
	jmp .6
.5:
	mov bx, _brk
	mov ax, bx
	add ax, # tsk_size
	mov _brk, ax
.6:
	mov [bx+tsk_zzz], # 0

	mov [bx+tsk_ss], cx
	mov [bx+tsk_sp], dx

	mov si, task            ; link new task after current one
	mov ax, [si+tsk_next]
	mov [bx+tsk_next], ax
	mov [si+tsk_next], bx

	and [bp+6], # ~1        ; clear carry
	mov dx, # _DS
	mov ax, bx

	pop ds
	pop si
	pop bp
	iret

;----------------------------------------------------------------------

	macro interrupt_epilog   ; optional eoi_cmd
	pop es
	pop ds
	if ?1                    ; eoi?
	mov ax, # ?1             ; EOI_xxx
	mov dx, # EOI
	cli
	out dx, PCB_AX
	endif
	popa
	iret
	mend

	macro interrupt_prolog
	pusha
	push ds
	push es                  ; must match IREGS definition
	mov ax, # _DS
	mov ds, ax
	mov es, ax
	cld
	mend


	if 0

	mov bx, task             ; XXX hmmm...
	cmp bx, # task_0
	jz already
	mov [bx+tsk_ss], ss
	mov [bx+tsk_sp], sp

	mov bx, # task_0
	mov task, bx
	mov ss, [bx+tsk_ss]
	mov sp, [bx+tsk_sp]
already:


	; or

	mov bx, ss
	cmp bx, ax          ; _DS
	je already
	mov [ss_save], ss
	mov [sp_save], sp
	mov ss, ax
	mov sp, # XXX
already:

	; and, in epilog

	cmp sp, # XXX
	jne nested
	mov bx, ss
	cmp bx, # _DS
	jne nested
	mov ss, [ss_save]
	mov sp, [sp_save]
nested:

	endif

;----------------------------------------------------------------------
; software interrupt 0x23
;
; no args, all regs kept

yield:                      ; interrupts left disabled throughout
	interrupt_prolog        ; IREGS to old stack

ret_next:
	mov bx, task
	mov [bx+tsk_ss], ss
	mov [bx+tsk_sp], sp

	mov bx, [bx+tsk_next]
ret_bx:
	mov task, bx
	mov ss, [bx+tsk_ss]
	mov sp, [bx+tsk_sp]

	interrupt_epilog 0      ; IREGS from new stack

;----------------------------------------------------------------------
; exceptions

	import _dump

	macro oops           ; descriptive string
	.data
msg set *
	.asciz ?1

	.text
	pusha
	mov ax, # msg
	jmp crash
	mend

crash:                   ; after pusha mov ax, # msg
	push ds
	push es
	mov bx, # _DS
	mov ds, bx
	mov es, bx
	cld

	push ss              ; 'ss' extra for dump()
	call _dump           ; 'msg' still in ax

	jmp near _reboot

floating_escape:
	oops "floating escape"

breakpoint:
	oops "breakpoint"

single_step:             ; XXX not worth trying to clear TF from stack
	oops "single step"

bounds_check:
	oops "bounds check"

overflow:
	oops "arithmetic overflow"

divide_error:
	oops "divide error"

invalid_opcode:
	oops "invalid opcode"

reserved:
	oops "reserved interrupt"

;----------------------------------------------------------------------

	macro interrupt          ; handler and optional eoi_cmd
	interrupt_prolog
	import ?1
	call ?1
	interrupt_epilog ?2
	mend

dma_0_int:
	sti
	interrupt _dma_0_int, EOI_DMA0

int_3:
	sti
	interrupt _int_3, EOI_INT3

timer_0_int:
	sti
	interrupt _timer_0_int, EOI_TMR

timer_1_int:
	sti
	interrupt _timer_1_int, EOI_TMR

;----------------------------------------------------------------------
; 5 uarts
;
; XXX priority a in asic ?

	import uart_0_int
	import uart_1_int
	import uart_2_int
	import uart_3_int
	import uart_4_int

int_0:
	sti
	interrupt_prolog

	in al, INT_0_REASON
again_0:
	mov bp, ax

	ror bp, # 1
	jnc not_0_0

	call uart_0_int
not_0_0:
	ror bp, # 1
	jnc not_0_1

	call uart_1_int
not_0_1:
	ror bp, # 1
	jnc not_0_2

	call uart_2_int
not_0_2:
	ror bp, # 1
	jnc not_0_3

	call uart_3_int
not_0_3:
	ror bp, # 1
	jnc not_0_4

	call uart_4_int
not_0_4:

	in al, INT_0_REASON
	or al, al
	jnz again_0

	interrupt_epilog EOI_INT0

;----------------------------------------------------------------------
; keyboard
; keypad + barcode reader
;
; XXX priority c in asic ?

	import _keyboard_rint
	import _keyboard_tint

int_1:
	sti
	interrupt_prolog

	in al, ASIC+0x1D             ; XXX status/interrupt enable reg
again_1:
	test al, # 2                 ; rint ?
	jz not_1_2

	push ax

	in al, ASIC+0x1A             ; data register
	call _keyboard_rint
	pop ax
not_1_2:
	test al, # 1                 ; XXX tint bit
	jz not_1_1

	call _keyboard_tint
not_1_1:
	in al, ASIC+0x1D             ; XXX status/interrupt enable reg
	test al, # 3                 ; r or tint ?
	jnz again_1

	interrupt_epilog EOI_INT1

;----------------------------------------------------------------------
; input pin monitor and timer
;
; XXX priority b in asic ?

	import _pa_chg
	import _pb_chg
	import _asic_timer_int
	import _netfree

int_2:
	sti
	interrupt_prolog

	in al, INT_2_REASON
again_2:
	mov bp, ax            ; XXX

	ror bp, # 1
	jnc not_2_0

	in ax, PA_DAT         ; PA_CHG to ah
still_2_0:
	call _pa_chg

	in ax, PA_DAT
	or ah, ah
	jne still_2_0         ; XXX no cure for the lockups. probably unnecessary

not_2_0:
	ror bp, # 1
	jnc not_2_1

	in ax, PB_DAT         ; PB_CHG to ah
still_2_1:
	call _pb_chg

	in ax, PB_DAT
	or ah, ah
	jne still_2_1         ; another one immediately

not_2_1:
	ror bp, # 1
	jnc not_2_2

	in al, ASIC+0x29      ; asic timer
	call _asic_timer_int  ; XXX similar cures needed here ?
not_2_2:
	ror bp, # 1
	jnc not_2_3

	in al, ASIC+0x18      ; MBUS (and ?) idle monitor
	call _netfree
not_2_3:

	in al, INT_2_REASON
	or al, al
	jnz again_2

	interrupt_epilog EOI_INT2

;----------------------------------------------------------------------
; 256 Hz periodic

timer_2_int:
	sti
	interrupt_prolog

	mov al, # 1
	out ASIC+0x2A, al     ; holds off watchdog in ASIC by 1 second

	dec _fido
	jz reboot_            ; mainline is stuck

	mov bx, # _timex      ; bump timex by 1/256 seconds
	add [bx+0], # 0x10000 / HZ
	adc [bx+2], # 0
	adc [bx+4], # 0

	mov ax, # 1           ; force dma_1_int at PR=7
	mov dx, # D1CNT
	out dx, PCB_AX

	mov ax, # 0x9306      ; Unsynchronized, TC INT, CHG STRT
	mov dx, # D1CON
	out dx, PCB_AX

	interrupt_epilog EOI_TMR

reboot_:
	jmp near _reboot

;----------------------------------------------------------------------

dma_1_int:
	sti
	interrupt_prolog

	call _tick

	interrupt_epilog EOI_DMA1

;----------------------------------------------------------------------
	export _disable, _restore, _enable
	export _spl_up, _splx
	export _inb, _outb
	export _inw, _outw_pcb

_disable:             ; word disable();
	pushf
	pop ax            ; oflags
	cli
	ret

_restore:             ; void restore(word oflags);
	and ah, # 0x02    ; ARG1. FLAGS.IF ?
	jz L4
_enable:              ; void enable();
	sti
L4:
	ret

_splx:                ; void splx(opri);
	mov dx, # IPRIO
	out dx, PCB_AX    ; ARG1
	ret

_spl_up:              ; word spl_up(npri);
	mov bx, ax        ; ARG1
	mov dx, # IPRIO
	in ax, dx         ; opri
	cmp ax, bx
	jbe .1
	xchg ax, bx
	out dx, PCB_AX
	xchg ax, bx
.1:
	ret

_inb:                 ; word inb(port);
	mov dx, ax        ; ARG1
	mov ah, # 0       ; zeroextend
	in al, dx
	ret

_inw:                 ; word inw(port);
	mov dx, ax        ; ARG1
	in ax, dx
	ret

_outb:                ; void outb(port, byte);
	mov dx, ax        ; ARG1
	ARG2 al
	out dx, al
	ret

_outw_pcb:            ; void outw_pcb(port, word);
	mov dx, ax        ; ARG1
	ARG2 ax
	out dx, PCB_AX
	ret

;----------------------------------------------------------------------

_dma:               ; void dma(int ch, word con, word cc, dword src, dword dst)
	push bp         ; 0  2  4   6  8    12
	mov bp, sp      ; bp ra con cc s:rc d:st

	or  ax, ax
	mov dx, # D0SRCL
	jz  dma_0
	mov dx, # D1SRCL
dma_0:
	mov ax, [bp+8]
	out dx, PCB_AX  ; DxSRCL
	inc dx
	inc dx
	mov ax, [bp+10]
	out dx, PCB_AX  ; DxSRCH
	inc dx
	inc dx
	mov ax, [bp+12]
	out dx, PCB_AX  ; DxDSTL
	inc dx
	inc dx
	mov ax, [bp+14]
	out dx, PCB_AX  ; DxDSTH
	inc dx
	inc dx
	mov ax, [bp+6]
	out dx, PCB_AX  ; DxCNT
	inc dx
	inc dx
	mov ax, [bp+4]
	out dx, PCB_AX  ; DxCON

	leave
	ret

;----------------------------------------------------------------------
; for delays shorter than 16 msec

	export _busywait_usec

_busywait_usec:
	mov bx, # 256 * TMRCLK / 1000000 ; usec to ticks of Timer 1
	mul bx
	mov bl, ah                       ; /= 256
	mov bh, dl
	mov dx, # T1CNT
	in  ax, dx
	add bx, ax                       ; target ticks
not_yet:
	in  ax, dx
	cmp bx, ax
	jns not_yet
	ret

;----------------------------------------------------------------------

_time:                ; dword time(dword *);
	mov bx, ax        ; timep or NULL
	mov cx, es
	les ax, [_timew]  ; atomic
	mov dx, es        ; dword return
	mov es, cx
	or  bx, bx
	jz  no_timep
	mov [bx+0], ax
	mov [bx+2], dx
no_timep:
	ret

;----------------------------------------------------------------------

