#if 0
	cc -fpack-struct mkfs.c -o mkfs
	exit
#endif

#define int short

#include "../h/param.h"
#include "../h/inode.h"
#include "../h/ino.h"
#include "../h/filsys.h"
#include "../h/fblk.h"
#include "../h/dir.h"

#undef int

#define WBLK(bno, what) wblk(bno, (void *) &(what), sizeof(what))
#define RBLK(bno, what) rblk(bno, (void *) &(what), sizeof(what))

static struct filsys filsys;
static ino_t ino_feed;
static char zeroes[BSIZE];

static char *fsname = "fs.out";
static int fso;

static void
wblk(daddr_t bno, char *buf, int len)
{
	lseek(fso, (long long) BSIZE * bno, 0);

	if (len > BSIZE) {
		printf("cannot write %d bytes to block %d\n", len, bno);
		exit(1);
	}
	if (write(fso, buf, len) != len) {
		perror(fsname);
		exit(1);
	}
	if (len < BSIZE) {
		if (write(fso, zeroes, BSIZE - len) != BSIZE - len) {
			perror(fsname);
			exit(1);
		}
	}
}

static void
rblk(daddr_t bno, char *buf, int len)
{
	int n;

	if (len > BSIZE) {
		printf("cannot read %d bytes from block %d\n", len, bno);
		exit(1);
	}
	lseek(fso, (long long) BSIZE * bno, 0);

	n = read(fso, buf, len);
	if (n != len) {
		perror(fsname);
		exit(1);
	}
}

static void
ltol3(char *di, char *ci, int n)
{
	do {
		*di++ = *ci++;
		*di++ = *ci++;
		*di++ = *ci++; ci++;
	} while (--n);
}

static void
bfree(daddr_t bno)
{
	int i;
	struct fblk fblk;

	if (filsys.s_nfree >= NICFREE) {
		bzero(&fblk, sizeof(fblk));
		fblk.df_nfree = filsys.s_nfree;
		for (i = 0; i < NICFREE; ++i)
			fblk.df_free[i] = filsys.s_free[i];
		WBLK(bno, fblk);

		filsys.s_nfree = 0;
	}
	filsys.s_free[filsys.s_nfree++] = bno;
}

static daddr_t
alloc()
{
	int i;
	daddr_t bno;
	struct fblk fblk;

	bno = filsys.s_free[--filsys.s_nfree];
	if (bno == 0) {
		printf("out of free space\n");
		exit(1);
	}
	if (filsys.s_nfree == 0) {
		RBLK(bno, fblk);
		for (i = 0; i < NICFREE; ++i)
			filsys.s_free[i] = fblk.df_free[i];

		filsys.s_nfree = fblk.df_nfree;
	}
	return (bno);
}

static void
newblk(char *db, int *adbc, int *aibc, daddr_t *ib)
{
	daddr_t bno;

	*adbc = 0;

	bno = alloc();
	wblk(bno, db, BSIZE);
	bzero(db, BSIZE);

	ib[*aibc] = bno;

	if (++*aibc >= NINDIR) {
		*aibc = 0;
		printf("indirect block full\n");
		exit(1);
	}
}

static void
entry(ino_t inum, char *name, char *db, int *adbc, int *aibc, daddr_t *ib)
{
	struct direct *dp = (struct direct *) db + *adbc;

	dp->d_ino = inum;
	strncpy(dp->d_name, name, DIRSIZ);

	if (++*adbc >= BSIZE / sizeof(*dp))
		newblk(db, adbc, aibc, ib);
}

static void
iput(struct inode *ip, int ibc, daddr_t *ib)
{
	struct dinode *dp;
	daddr_t d;
	int i;
	char buf[BSIZE];

	d = itod(ip->i_number);
	if (d >= filsys.s_isize) {
		printf("overflow inode %d, blk %d\n", ip->i_number, d);
		exit(1);
	}
	RBLK(d, buf);
	dp = (struct dinode *) buf + itoo(ip->i_number);

	dp->di_mode  = ip->i_mode;
	dp->di_nlink = ip->i_nlink;
	dp->di_size  = ip->i_size;

#define LADDR (NADDR-3)

	switch (ip->i_mode & IFMT) {

	case IFDIR:
	case IFREG:
		for (i = 0; i < ibc; i++) {
			if (i >= LADDR)
				break;
			ip->i_addr[i] = ib[i];
		}
		if (ibc >= LADDR) {
			ip->i_addr[LADDR] = alloc();

			wblk(ip->i_addr[LADDR], ib + i, (ibc - i) * sizeof(*ib));
		}

	case IFBLK:
	case IFCHR:
		ltol3(dp->di_addr, (char *) ip->i_addr, NADDR);
		break;

	default:
		printf("bad mode 0%o\n", ip->i_mode);
		exit(1);
	}
	WBLK(d, buf);
}

static void
bflist()
{
	daddr_t d;
	struct inode in;

	bfree(0);

	d = filsys.s_fsize;
	while (--d >= filsys.s_isize)
		bfree(d);

	/* inode 1 allocates bad blocks */

	bzero(&in, sizeof(in));

	in.i_number = 1;
	in.i_nlink = 0;
	in.i_mode = IFREG;

	iput(&in, 0, NULL);
}

#if 0

/usr/mdec/uboot
4872 55
drwx
usr     drwx
        sh      -rwx /bin/sh
        ken     drwx
                $
        b0      brw- 0 0
        c0      crw- 0 0
        $
$

#endif

main(int argc, char **argv)
{
	int x;

	if (argc != 3) {
		printf("Usage: mkfs fsname proto\n");
		exit(2);
	}

	fsname = argv[1];

	close(0);
	if (open(argv[2], 0) != 0) {
		perror(argv[2]);
		exit(1);
	}

	/* bootfile
	 * blocks inodes */

	scanf("%*s");

	scanf("%d", &x);
	filsys.s_fsize  = x;

	scanf("%d", &x);
	filsys.s_isize  = (x + INOPB-1) / INOPB + SUPERB+1;

	printf("%s %d blocks, %d inodes, %d kB data\n",
		fsname,
		filsys.s_fsize,
		(filsys.s_isize - (SUPERB+1)) * INOPB - ROOTINO,
		(filsys.s_fsize - filsys.s_isize) * BSIZE / 1024);

	fso = open(fsname, 2, 0666);
	if (fso == -1) {
		perror(fsname);
		exit(1);
	}

	for (x = SUPERB+1; x < filsys.s_isize; ++x)
		wblk(x, zeroes, BSIZE);

	bflist();

	ino_feed = ROOTINO;
	cfile(0);

	WBLK(SUPERB, filsys);
	exit(0);
}

cfile(par)
	struct inode *par;
{
	int dbc, ibc;
	int fd, cc;
	int maj, min;
	struct inode in;
	char string[256];
	daddr_t ib[NINDIR];
	char db[BSIZE];

	dbc = 0;
	bzero(db, sizeof(db));
	ibc = 0;
	bzero(ib, sizeof(ib));

	/*
	 * dr-x
	 * -r-- sourcefile
	 * brw- 0,0
	 * crw- 2,0
	 */

	bzero(&in, sizeof(in));

	in.i_number = ino_feed++;
	in.i_nlink = 1;

	if (par == 0) {
		par = &in;
		in.i_nlink--;
	}

	scanf("%s", string);
	if (string[1] == 'r') in.i_mode |= 4;
	if (string[2] == 'w') in.i_mode |= 2;
	if (string[3] == 'x') in.i_mode |= 1;

	switch (string[0]) {

	case 'b':
		in.i_mode = IFBLK;
		scanf("%d,%d", &maj, &min);
		in.i_addr[0] = (maj<<8) | min;
		break;

	case 'c':
		in.i_mode = IFCHR;
		scanf("%d,%d", &maj, &min);
		in.i_addr[0] = (maj<<8) | min;
		break;

	case 'd':
		in.i_mode = IFDIR;
		in.i_nlink++;
		par->i_nlink++;

		entry(in.i_number,   ".",  db, &dbc, &ibc, ib);
		entry(par->i_number, "..", db, &dbc, &ibc, ib);

		in.i_size = 2 * sizeof(struct direct);
		for (;;) {
			string[0] = '$';
			scanf("%s", string);
			if (string[0] == '$')
				break;

			in.i_size += sizeof(struct direct);

			entry(ino_feed, string, db, &dbc, &ibc, ib);
			cfile(&in);
		}
		break;

	default:
		in.i_mode = IFREG;

		scanf("%s", string);
		fd = open(string, 0);
		if (fd == -1) {
			perror(string);
			exit(1);
		}
		while ((cc = read(fd, db, BSIZE)) > 0) {
			in.i_size += cc;
			newblk(db, &dbc, &ibc, ib);
		}
		close(fd);
		break;
	}
	if (dbc)
		newblk(db, &dbc, &ibc, ib);

	iput(&in, ibc, ib);
}

