/*
 * The I node is the focus of all
 * file activity in unix. There is a unique
 * inode allocated for each active file,
 * each current directory, each mounted-on
 * file, text file, and the root. An inode is 'named'
 * by its dev/inumber pair. (iget/iget.c)
 * Data, from mode on, is read in
 * from permanent inode on volume.
 */

#define	NADDR	19

struct	inode
{
	char	i_flag;
	char	i_count;	/* reference count */
	ino_t	i_number;	/* i number, 1-to-1 with device address */
	struct	mount *i_mp;/* device where inode resides */
	daddr_t	i_lastr;	/* last logical block read (for read-ahead) */
	off_t	i_size;		/* size of file */
	char	i_nlink;	/* directory entries */
	unsigned char i_mode;
	daddr_t	i_addr[NADDR];	/* if normal file/directory */
};

#define i_rdev   i_addr[0]
#define LINK_MAX 255

extern	struct inode inode[];	/* The inode table itself */

extern struct	inode *rootdir;		/* pointer to inode of root directory */

struct	inode *ialloc();
struct	inode *iget();
struct	inode *owner();
struct	inode *maknode();
struct	inode *namei();

/* flags */
#define	ILOCK	01		/* inode is locked */
#define	IWANT	02		/* some process waiting on lock */
#define	IMOUNT	04		/* inode is mounted on */
#define	ITEXT	010		/* inode is pure text prototype */
#define	IUPD	020		/* file has been modified */
#define	ICHG	040		/* inode has been changed */

/* modes */
#define	IFMT	0xE0	/* type of file, 20 set if any kind of special */
#define	IFREG	0x80	/* regular */
#define	IFDIR	0xC0	/* directory */
#define	IFBLK	0x60	/* block special */
#define	IFCHR	0x20	/* character special */

#define	IPERM	0x07	/* read, write, execute permissions */
#define	IREAD	0x04
#define	IWRITE	0x02
#define	IEXEC	0x01

