/*
 * A clist structure is the head
 * of a linked list queue of characters.
 * The characters are stored in 4-word
 * blocks containing a link and several characters.
 * The routines getc and putc
 * manipulate these structures.
 */
struct clist {
	int	c_cc;		/* character count */
	char	*c_cf;		/* pointer to first char */
	char	*c_cl;		/* pointer to last char */
};

/*
 * A tty structure is needed for
 * each UNIX character device that
 * is used for normal terminal IO.
 * The routines in tty.c handle the
 * common code associated with
 * these structures.
 * The definition and device dependent
 * code is in each driver. (kl.c dz.c)
 */

struct tty {
	struct	clist t_rawq;	/* input chars right off device */
	struct	clist t_canq;	/* input chars after erase and kill */
	struct	clist t_outq;	/* output list to device */
	int	(*t_oproc)();	/* routine to start/stop output */
	caddr_t	t_addr;		/* device address */
	short	t_state;	/* internal state, not visible externally */
	short	t_pgrp;		/* process group name */
	char	t_flags;	/* mode, settable by ioctl call */
	char	t_speed;	/* input/output speed */
	char	t_erase;	/* erase character */
	char	t_kill;		/* kill character */
	char	t_intrc;	/* interrupt */
	char	t_delct;	/* number of delimiters in raw q */
	char	t_col;		/* printing column of device */
};

/*
 * structure of arg for ioctl
 */
struct	ttiocb {
	char	ioc_speed;
	char	ioc_erase;
	char	ioc_kill;
	char	ioc_flags;
};

#define	TTIPRI	28
#define	TTOPRI	29

#define	CEOT	004		/* ^D (hardcoded) */
#define	CERASE	010		/* ^H */
#define	CSTART	021		/* ^Q (hardcoded) */
#define	CSTOP	023		/* ^S (hardcoded) */
#define	CKILL	025		/* ^U */
#define	CINTR	0177	/* DEL */

/* limits */
#define	TTHIWAT	650
#define	TTLOWAT	125
#define	TTYHOG	256

/* t_flags */
#define	TANDEM	01
#define	CBREAK	02
#define	ECHO	04
#define	CRMOD	010
#define	RAW		020
#define	ODDP	040
#define	EVENP	0100
#define	XTABS	0200

/* t_state */
#define	WOPEN	01		/* Waiting for open to complete */
#define	ISOPEN	02		/* Device is open */
#define	CARR_ON	04		/* Software copy of carrier-present */
#define	BUSY	010		/* Output in progress */
#define	TIMEOUT	020		/* Delay timeout in progress */
#define	ASLEEP	040		/* Wakeup when output done */
#define	TTSTOP	0100	/* Output stopped by ctl-s */
#define	TBLOCK	0200	/* tandem queue blocked */
#define	FLUSH	0400	/* outq has been flushed during DMA */
#define	XCLUDE	01000	/* exclusive-use flag against open */
#define	HUPCLS	02000	/* Hang up upon last close */

