
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/buf.h"
#include "../h/callo.h"

#define	SCHMAG	9/10

struct callout *calltodo, *callfree, callspace[NCALL];
char recalc;

time_t time;

hardclock()
{
	register struct callout *c;
	register struct proc *p;

	for (c = calltodo; c; c = c->c_next)
		if (c->c_time) {
			c->c_time--;
			break;
		}

	p = curproc;
	if (p->p_stat == SRUN && p->p_cpu != 255)
		++p->p_cpu;

	if (++lbolt >= HZ) {
		lbolt -= HZ;
		++time;
		runrun = 1;
		recalc = 1;
	}
}

softclock()
{
	register struct callout *c;
	register struct proc *p;
	void (*fun)();
	unsigned arg;
	int a;

	splclock();
	while (c = calltodo) {
		if (c->c_time)
			break;

		calltodo = c->c_next;
		c->c_next = callfree;
		callfree = c;

		fun = c->c_func;
		arg = c->c_arg;

		spl0();
		(*fun)(arg);
		splclock();
	}
	spl0();

	if (recalc) {
		recalc = 0;
		wakeup(&lbolt);

		for (p = proc; p < &proc[NPROC]; ++p) {
			if (p->p_stat > SIDL) {

				a = (p->p_cpu & 255)*SCHMAG + p->p_nice - NZERO;
				if (a < 0)
					a = 0;
				if (a > 255)
					a = 255;
				p->p_cpu = a;

				if (p->p_pri >= PUSER)
					setpri(p);

				if (p->p_clktim && --p->p_clktim == 0)
					psignal(p, SIGALRM);
			}
		}
	}
}

/*
 * timeout is called to arrange that
 * fun(arg) is called in tim/HZ seconds.
 * An entry is sorted into the callout
 * structure. The time in each structure
 * entry is the number of HZ's more
 * than the previous entry.
 * In this way, decrementing the
 * first entry has the effect of
 * updating all entries.
 */
timeout(fun, arg, tim)
	void (*fun)();
	unsigned arg, tim;
{
	register struct callout *c, **pc = &calltodo;
	struct callout *nc;
	int s;

	s = splclock();
	nc = callfree;
	callfree = nc->c_next;
	if (nc == NULL)
		panic("Out of callouts");

	while (c = *pc) {
		if (c->c_time > tim) {
			c->c_time -= tim;
			break;
		}
		tim -= c->c_time;
		pc = &c->c_next;
	}
	nc->c_next = c;
	nc->c_time = tim;
	nc->c_arg  = arg;
	nc->c_func = fun;
	*pc = nc;
	splx(s);
}

void
callout_init()
{
	register struct callout *c = callspace;

	do {
		c->c_next = callfree;
		callfree = c;
	} while (++c != &callspace[NCALL]);
}

