
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/inode.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/buf.h"

/*
 * Convert a pathname into a pointer to
 * an inode. Note that the inode is locked.
 *
 * flag:
 */

enum { LOOKUP, CREATE, DELETE };

struct inode *
namei(fname, flag)
	char *fname;
{
	register struct inode *dp;
	register c;
	register char *cp;
	struct direct *dent = &u.u_dent;
	struct buf *bp;
	struct mount *mp;
	off_t eo;

	/*
	 * If name starts with '/' start from root;
	 * otherwise start from current dir.
	 */

	dp = curproc->p_cdir;
	if ((c = fubyte(fname++)) == '/')
		dp = rootdir;

	while (c == '/')
		c = fubyte(fname++);

	if (c == 0 && flag != LOOKUP) {
		errno = EINVAL;
		return (NULL);
	}

	VOID iget(dp->i_mp, dp->i_number);

cloop:
	/*
	 * Here dp contains pointer to last component matched.
	 */

	if (errno)
		goto out;

	if (c == 0)
		return (dp);

	/*
	 * If there is another component,
	 * Gather up name into users' dir buffer.
	 */

	cp = u.u_dbuf;
	bzero(cp, DIRSIZ);

	while (c != '/' && c > 0) {
		if (cp < &u.u_dbuf[DIRSIZ])
			*cp++ = c;
		c = fubyte(fname++);
	}
	while (c == '/')
		c = fubyte(fname++);

seloop:
	/*
	 * dp must be a directory
	 */

	if ((dp->i_mode&IFMT) != IFDIR)
		errno = ENOTDIR;

	if (errno)
		goto out;

	/*
	 * set up to search a directory
	 */
	u.u_offset = 0;
	u.u_segflg = 0; /* XXX why is this here ? */
	eo = 0;
	bp = NULL;

eloop:

	/*
	 * If at the end of the directory, the search failed.
	 * Report what is appropriate as per flag.
	 */

	if (u.u_offset >= dp->i_size) {
		if (bp)
			brelse(bp);

		if (flag == CREATE && c == 0) {
			if (!access(dp, IWRITE))
				goto out;

			if (eo)
				u.u_offset = eo - sizeof(struct direct);
			else
				dp->i_flag |= IUPD|ICHG;

			u.u_pdir = dp;
			return (NULL);
		}
		errno = ENOENT;
		goto out;
	}

	/*
	 * If offset is on a block boundary, read the next directory block.
	 * Release previous if it exists.
	 */

	if ((u.u_offset&BMASK) == 0) {
		if (bp)
			brelse(bp);
		bp = bread(dp->i_mp->m_dev, bmap(dp, (daddr_t)(u.u_offset>>BSHIFT), B_READ));
		if (bp->b_flags & B_ERROR) {
			brelse(bp);
			goto out;
		}
	}

	/*
	 * Note first empty directory slot in eo for possible creat.
	 * String compare the directory entry and the current component.
	 * If they do not match, go back to eloop.
	 */

	bcopy(bp->b_un.b_addr+(int)(u.u_offset&BMASK), dent, sizeof(*dent));
	u.u_offset += sizeof(*dent);

	if (dent->d_ino == 0) {
		if (eo == 0)
			eo = u.u_offset;
		goto eloop;
	}
	if (strncmp(dent->d_name, u.u_dbuf, DIRSIZ) != 0)
		goto eloop;

	/*
	 * Here a component matched in a directory.
	 * If there is more pathname, go back to cloop, otherwise return.
	 */

	if (bp)
		brelse(bp);

	if (flag == DELETE && c == 0) {
		if (!access(dp, IWRITE))
			goto out;

		return (dp);
	}
	mp = dp->i_mp;

	if (dp->i_number == ROOTINO
	 && dent->d_ino == ROOTINO
	 && dent->d_name[1] == '.') {  /* differentiate . and .. in root */
		iput(dp);
		dp = mp->m_mpoint;         /* step into parent fs */
		dp->i_count++;
		plock(dp);
		goto seloop;
	}
	iput(dp);
	dp = iget(mp, dent->d_ino);
	if (dp == NULL)
		return (NULL);

	goto cloop;

out:
	iput(dp);

	return (NULL);
}

