
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/inode.h"

int curpri;

#define SQSIZE 64	/* power of 2 */

#define HASH(x)	(((unsigned) x >> 5) & (SQSIZE-1))

struct proc
	*runq,
	*slpque[SQSIZE];

/*
 * Give up the processor till a wakeup occurs
 * on chan, at which time the process
 * enters the scheduling queue at priority pri.
 * The most important effect of pri is that when
 * pri<=PZERO a signal cannot disturb the sleep;
 * if pri>PZERO signals will be processed.
 * Callers of this routine must be prepared for
 * premature return, and check that the reason for
 * sleeping has gone away.
 */
sleep(chan, pri)
	caddr_t chan;
{
	register struct proc *p, **que;
	int s;

	if (chan == 0)
		panic("zero wchan");

	que = &slpque[HASH(chan)];
	p = curproc;

	s = splhigh();
	if (pri > PZERO) {
		if (issig()) {
			splx(s);
			return (1);
		}
	}
	p->p_wchan = chan;
	p->p_pri   = pri;
	p->p_stat  = SSLEEP;
	p->p_link  = *que;
	*que = p;

	swtch();
	splx(s);
	return (p->p_wchan != NULL);
}

/*
 * Wake up all processes sleeping on chan.
 */
wakeup(chan)
	register caddr_t chan;
{
	register struct proc *p, **pp;
	int s;

	pp = &slpque[HASH(chan)];

	s = splhigh();

	while (p = *pp) {
		if (p->p_wchan == chan) {
			p->p_wchan = NULL;
			*pp = p->p_link;
			setrun(p);
		} else
			pp = &p->p_link;
	}
	splx(s);
}

static void
setrq(p)
	struct proc *p;
{
	register s = splhigh();

	p->p_link = runq;
	runq = p;
	splx(s);
}

/*
 * Set the process running
 */
setrun(p)
	register struct proc *p;
{
	register struct proc **pp;
	int s;

	if (p->p_stat <= SZOMB)
		panic("setrun zomb");

	/*
	 * sleeping ?
	 */
	s = splhigh();
	if (p->p_stat == SSLEEP) {
		for (pp = &slpque[HASH(p->p_wchan)]; *pp != p; pp = &(*pp)->p_link)
			if (*pp == NULL)
				panic("setrun slpque");

		*pp = p->p_link;
	}
	p->p_stat = SRUN;
	p->p_link = runq;
	runq = p;
	splx(s);
	if (p->p_pri < curpri)
		runrun = 1;
}

/*
 * Set user priority.
 * The rescheduling flag (runrun)
 * is set if the priority is better
 * than the currently running process.
 */
setpri(p)
	register struct proc *p;
{
	register pri;

	pri = (p->p_cpu & 0377) / 16 + p->p_nice + PUSER - NZERO;
	if (pri > 127)
		pri = 127;
	p->p_pri = pri;
	if (pri < curpri)
		runrun = 1;

	return (pri);
}

/*
 * put the current process on
 * the Q of running processes and
 * call the scheduler.
 */
qswtch()
{
	setrq(curproc);
	swtch();
}

/*
 * This routine is called to reschedule the CPU.
 * if the calling process is not in RUN state,
 * arrangements for it to restart must have
 * been made elsewhere, usually by calling via sleep.
 */
swtch()
{
	register struct proc *p, **pp, **hp;
	register pri;

	/*
	 * Search for highest-priority runnable process.
	 */
	pri = 128;
	hp = NULL;

	for (;;) {
		splhigh();

		for (pp = &runq; (p = *pp); pp = &p->p_link)
			if (p->p_stat == SRUN && (p->p_flag&SLOAD)) {
				if (pri > p->p_pri) {
					pri = p->p_pri;
					hp = pp;
				}
			}

		if (hp)
			break;
		idle();
	}
	p = *hp;
	*hp = p->p_link;

	curpri = pri;
	runrun = 0;
	spl0();

	asm_switch(p);
}

