
#include <sys/types.h>
#include <sys/soundcard.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

static int
check_crc(u_char *p, int n)
{
	unsigned x = 0xFFFF;

	do {
		int b = 8;

		x ^= *p++;
		do {
			int ox = x;

			x >>= 1;
			if (ox & 1)
				x ^= 0x8408;
		} while (--b);
	} while (--n);

	return (((p[1] << 8) | p[0]) == (~x & 0xFFFF));
}

typedef short sample_t;

#define SAMPLERATE 44100

static void
audio(sample_t sample)
{
	static sample_t was;
	static unsigned tone, ICR_zero, ICR_2, ICR_1, ICR;

	static unsigned b, bc, cc;

	unsigned t;

	ICR++;

	/* zero crossing ? */

	if (((sample ^ was) & 0x8000) == 0)
		return;

	was = sample;

	ICR_2 = ICR_1;
	ICR_1 = ICR;
	t = ICR_1 - ICR_2;

	/* below or above 1700 Hz ? */

	t -= 13;
	if (((t ^ tone) & 0x8000) == 0)
		return;

	tone = t;

	t = ICR_2 - ICR_zero;
	ICR_zero = ICR_2;

	/* how many bittimes ? */

	switch ((t + 36 / 2) / 36) {
	default:
		if (cc >= 15)
			printf("\n");
		cc = bc = 0;
		return;

	case 6:
		bc += 5;
		b >>= 5;
		b |= 0xF800;
		break;
	case 5:
		bc += 5;
		b >>= 5;
		b |= 0x7800;
		break;
	case 4:
		bc += 4;
		b >>= 4;
		b |= 0x7000;
		break;
	case 3:
		bc += 3;
		b >>= 3;
		b |= 0x6000;
		break;
	case 2:
		bc += 2;
		b >>= 2;
		b |= 0x4000;
		break;
	case 1:
	case 0:
		bc += 1;
		b >>= 1;
		break;
	}
	if (bc >= 8) {
		bc -= 8;
		if (cc < 14)
			printf("%c", (b >> (9 - bc)) & 0x7F);
		else
			printf("%c", b >> (8 - bc));
		cc++;
	}
}

main(int argc, char **argv)
{
	char *dev = (argc > 1) ? argv[1] : "/dev/dsp";
	int fd, n, try;

	fd = open(dev, O_RDONLY);
	if (fd == -1)
		goto perr;

	if (ioctl(fd, SNDCTL_DSP_RESET, NULL) == 0) {

		n = AFMT_S16_LE; /* XXX use _NE */
		try = n;
		if (ioctl(fd, SNDCTL_DSP_SETFMT, &try))
			goto perr;
		if (try != n) {
			fprintf(stderr, "%s cannot do AFMT_%s\n",
				dev, (n == AFMT_U8) ? "U8" : "S16");
			goto err;
		}

		n = 0;
		try = n;
		if (ioctl(fd, SNDCTL_DSP_STEREO, &try))
			goto perr;
		if (try != n) {
			fprintf(stderr, "%s cannot do %s\n", dev, n ? "stereo" : "mono");
			goto err;
		}

		n = SAMPLERATE;
		try = n;
		if (ioctl(fd, SNDCTL_DSP_SPEED, &try))
			goto perr;
		if (abs(try - n) > 0.01 * n) {
			fprintf(stderr, "%s could do %d, but not %d Hz\n", dev, try, n);
			goto err;
		}

		if (try != n)
			fprintf(stderr, "%d Hz error in samplerate ignored\n", try - n);

		ioctl(fd, SNDCTL_DSP_SYNC, NULL);
	}

	for (;;) {
		int
			n;
		sample_t
			*p,
			buf[SAMPLERATE / 10];

		n = read(fd, buf, sizeof(buf));

		if (n < sizeof(sample_t))
			break;

		p = buf;
		do {
			audio(*p++);
		} while ((n -= sizeof(sample_t)) > 0);
	}
	if (n)
perr:
	perror(dev);
err:
	exit(1);
}

