/*
 *  cold reset from 0xFFFF:0 (last 16 bytes of rom)
 *  fixes UMCS for 256 kB rom and
 *  jumps to here, boot: in 0xC000:0x0400
 */

#include "head.h"
#include "pcb.h"
#include "asic.h"
#include "syscalls.h"

#asm
	include asm.inc

	export _version
	export _timex   ; seconds, in different formats
	export _timew
	export _timef
	export _fido

;----------------------------------------------------------------------
	.data
__data:
_version:
	.asciz VERSION

INT_STK equ 1024    ; 9 nested, worst case
SYS_STK equ 512     ; one per task

	.space INT_STK  ; has to reside below all system stacks
intstack:
	.space 4        ; for saved sp,ss

_brk:
	.long __end

_farbrk:
	.word 0x1000

_bank_1:
	.word 0         ; bitmap of ram banks, 1 if piece in use
_bank_2:
	.word 0

_romseg:
	.word 0xC000    ; 8000, C000 or E000

astflag:
	.word 0
runrun:
	.word 0

	export _task

_task:
	.word task_0

_ready:
	.word task_0

free:
	.word 0

tsk_next  equ 0     ; XXX task.h
tsk_prev  equ 2
tsk_dsize equ 4     ; long
tsk_bound equ 8     ; two words
tsk_bank  equ 12
tsk_sp    equ 14
tsk_ss    equ 16
sys_fl    equ 18
sys_ip    equ 20
sys_sp    equ 22
sys_bp    equ 24
tsk_size  equ 32

	.space SYS_STK  ; just below task struct
task_0:
	.word task_0    ; next.   initially only task_0
	.word task_0    ; prev
	.long 0x10000   ; dsize   64kB
	.word 1,0xFFFE  ; bound
	.word 0         ; bank
	.word 0,0       ; sp,ss
	.space tsk_size ; XXX few bytes lost

_boottime:
	.word 0
	.long 0x386D2760 ; difference of timex and UTC

;----------------------------------------------------------------------
	.bss

_timex:             ; 32.16
	.space 1
_timef:             ; 8.8
	.space 1
_timew:             ; 16.0
_time:              ; 32.0
	.space 4

_fido:
	.space 2        ; watchdog downcount

	.comm junk, 2   ; safe location for DMA pointers

;----------------------------------------------------------------------
	.text
__text:

	entry boot
	export _reboot

;----------------------------------------------------------------------
; 256 interrupt vectors at 0:0
; all handlers currently in segment 0 (valid after first movsb, below)

	macro ivec
	.word ?1, 0            ; segment 0
	mend

	ivec divide_error      ; 00 0000
	ivec single_step       ; 01 0004
v_nmi:
	ivec nmi               ; 02 0008
	ivec breakpoint        ; 03 000C
	ivec overflow          ; 04 0010
	ivec bounds_check      ; 05 0014
	ivec invalid_opcode    ; 06 0018
	ivec floating_escape   ; 07 001C
	ivec timer_0_int       ; 08 0020
	ivec reserved          ; 09 0024
	ivec dma_0_int         ; 0A 0028
	ivec dma_1_int         ; 0B 002C
	ivec int_0             ; 0C 0030
	ivec int_1             ; 0D 0034
	ivec int_2             ; 0E 0038
	ivec int_3             ; 0F 003C
	ivec numerics          ; 10 0040
	ivec reserved          ; 11 0044
	ivec timer_1_int       ; 12 0048
	ivec timer_2_int       ; 13 004C
	ivec reserved          ; 14 0050
	ivec reserved          ; 15 0054
	ivec reserved          ; 16 0058
	ivec reserved          ; 17 005C
	ivec reserved          ; 18 0060
	ivec reserved          ; 19 0064
	ivec reserved          ; 1A 0068
	ivec reserved          ; 1B 006C
	ivec reserved          ; 1C 0070
	ivec reserved          ; 1D 0074
	ivec reserved          ; 1E 0078
	ivec reserved          ; 1F 007C last reserved

	ivec syscall           ; 20 0080 available for user
	ivec unknown           ; 21 0084
	ivec unknown           ; 22 0088
	ivec unknown           ; 23 008C
	ivec unknown           ; 24 0090
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown           ; 2F 00BC

	macro eight_unknown_ivecs
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	mend

	eight_unknown_ivecs    ; 30 00C0
	eight_unknown_ivecs    ; 38 00E0
	eight_unknown_ivecs    ; 40 0100
	eight_unknown_ivecs    ; 48 0120
	eight_unknown_ivecs    ; 50 0140
	eight_unknown_ivecs    ; 58 0160
	eight_unknown_ivecs    ; 60 0180
	eight_unknown_ivecs    ; 68 01A0
	eight_unknown_ivecs    ; 70 01C0
	eight_unknown_ivecs    ; 78 01E0
	eight_unknown_ivecs    ; 80 0200
	eight_unknown_ivecs    ; 88 0220
	eight_unknown_ivecs    ; 90 0240
	eight_unknown_ivecs    ; 98 0260
	eight_unknown_ivecs    ; A0 0280
	eight_unknown_ivecs    ; A8 02A0
	eight_unknown_ivecs    ; B0 02C0
	eight_unknown_ivecs    ; B8 02E0
	eight_unknown_ivecs    ; C0 0300
	eight_unknown_ivecs    ; C8 0320
	eight_unknown_ivecs    ; D0 0340
	eight_unknown_ivecs    ; D8 0360
	eight_unknown_ivecs    ; E0 0380
	eight_unknown_ivecs    ; E8 03A0
	eight_unknown_ivecs    ; F0 03C0
	eight_unknown_ivecs    ; F8 03E0

	;-------------------------------------------------
	; reset fixes SS:SP to 0:0,
	; UCS to 256 kB, C0000 ... FFFFF. R=101: 1 waitstate, no RDY,
	; and jumps to C000:0400 (boot)
	;
	; things should be straightened out in less than
	; one second, watchdog nmi.

boot:                  ; absolute address in ROM C000:0400
retry:                 ; relative jumps from below
	cli
	cld

	;-------------------------------------------------
	; memory and io chipselects

	; PCS0 at 0000, R=111: 3 waitstates, no RDY, for PCS0-3

	mov ax, # 0x003F
	mov dx, # PACS
	out dx, PCB_AX

	; MCS0 at 00000, R=001: 1 waitstate, external RDY, for MCS0-3

	mov ax, # 0x01F9        ; XXX external RDY must be used
	mov dx, # MMCS
	out dx, PCB_AX

	; MCS0-3: four times 128 kB = 512 kB
	; EX,MS=10: all PCS0-6 in i/o space
	; R=010: 2 waitstates, external RDY for PCS4-6 XXX

	mov ax, # 0xC0BA
	mov dx, # MPCS
	out dx, PCB_AX

	; MCS0 is at zero, so LCS is not/can not be used.
	; LMCS is not programmed

	;-------------------------------------------------
	; relocate from    C000:0 to ram,
	; text/data/bss/stack 0:0 64 kB

	mov bx, # __edata
	xor ax, ax         ; constants # __edata and # 0

	mov ss, ax         ; 0 final SS:SP
	mov sp, ax         ; 0 wraps to ss:0xFFFE on first push
	mov bp, ax

	mov es, ax         ; 0
	mov di, ax         ; 0 dst 0:0

	mov dx, cs         ; ROM: 0x8000, 0xC000 or 0xE000
	mov ds, dx
	mov si, ax         ; 0 src 0xC000:0

	mov cx, bx         ; __edata
	rep
	movsb              ; copy text/data. interrupt vectors (nmi) now valid.

	mov si, ax         ; rewind to 0
	mov di, ax

	mov cx, bx         ; __edata
	rep
	cmpsb              ; verify text/data
	jnz retry          ; while bad copy (Vcc coming up?)

	mov cx, di         ; end of initialized data
	neg cx             ; to end of segment
	rep
	stosb              ; zero bss and stack

	mov di, bx         ; rewind to __edata

	mov cx, di
	neg cx
	rep
	scasb              ; verify zero
	jnz retry          ; while bad copy

	mov ds, ax         ; 0 final DS and ES
	mov es, ax

	mov [_romseg], cs  ; ROM is all but forgotten from now on.

	jmp far 0:L1       ; 0 final CS
L1:
	;-------------------------------------------------
	; interrupt controller

	mov ax, # 0xFFFF   ; opened by each MSK=0 below
	mov dx, # IMASK
	out dx, PCB_AX

	mov ax, # 7        ; no levels masked
	mov dx, # IPRIO
	out dx, PCB_AX

	mov cx, # 256      ; flush pending interrupts XXX meaningless
	mov ax, # EOI_ANY
	mov dx, # EOI
L2:
	out dx, PCB_AX
	loop L2

	;-------------------------------------------------
	; external interrupts
	; uarts etc from asic
	; XXX edge vs level ?

	mov ax, # 0x10    ; level-trigger MSK=0 PR=0
	mov dx, # I0CON
	out dx, PCB_AX

	mov ax, # 0x13    ; level-trigger MSK=0 PR=3
	mov dx, # I1CON
	out dx, PCB_AX

	mov ax, # 0x11    ; level-trigger MSK=0 PR=1
	mov dx, # I2CON
	out dx, PCB_AX

	mov ax, # 0x14    ; level-trigger MSK=0 PR=4
	mov dx, # I3CON
	out dx, PCB_AX

	;-------------------------------------------------
	; Timer 1
	; freerunning, about 2.032 MHz, period about 32 msec

	mov ax, # 0xC001  ; EN /INH CONT
	mov dx, # T1MODE
	out dx, PCB_AX

	mov ax, # 0       ; 0x10000
	mov dx, # T1CMP
	out dx, PCB_AX

	;-------------------------------------------------
	; Timer 2
	; periodic interval timer, HZ

	mov ax, # 0x6001  ; /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	mov ax, # (TMRCLK + HZ/2) / HZ    ; should be close enough
	mov dx, # T2CMP
	out dx, PCB_AX

	mov ax, # 0
	mov dx, # T2CNT
	out dx, PCB_AX

	mov ax, # 0xE001  ; EN /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	;-------------------------------------------------
	; Timers 0, 1 and 2

	mov ax, # 2       ; MSK=0 PR=2
	mov dx, # TICON
	out dx, PCB_AX

	;-------------------------------------------------
	; DMA controller, not used

	mov bx, # 0
	mov ax, # junk         ; &junk

	mov dx, # D0SRCL
	out dx, PCB_AX
	mov dx, # D0DSTL
	out dx, PCB_AX
	mov dx, # D1SRCL
	out dx, PCB_AX
	mov dx, # D1DSTL
	out dx, PCB_AX
	mov ax, bx
	mov dx, # D0SRCH
	out dx, PCB_AX
	mov dx, # D0DSTH
	out dx, PCB_AX
	mov dx, # D1SRCH
	out dx, PCB_AX
	mov dx, # D1DSTH
	out dx, PCB_AX         ; &junk in all pointers

	mov ax, # 1
	mov dx, # D0CNT
	out dx, PCB_AX
	mov dx, # D1CNT
	out dx, PCB_AX

	; memory to memory, constant pointers, TC INT, byte xfer

	mov ax, # 0x9370      ; Timer 2, P
	mov dx, # D0CON
	out dx, PCB_AX

	mov ax, # 0x9300      ; Unsynchronized
	mov dx, # D1CON
	out dx, PCB_AX

	mov ax, # 8 | 0       ; MSK=1 PR=0
	mov dx, # D0ICON
	out dx, PCB_AX

	mov ax, # 0 | 7       ; MSK=0 PR=7
	mov dx, # D1ICON
	out dx, PCB_AX

	;-------------------------------------------------

	xor bp, bp            ; marks end of chain

	call _initialise
	sti

	call _start
	call _idle
	jmp _reboot

;----------------------------------------------------------------------

	jmp far 0xFFFF:0  ; the cold start. not used

;----------------------------------------------------------------------
; reasons:
;  power off, time left about 35 msec (if large backlight off)
;  1 second since last watchdog rewind, time left to reset 1 second

nmi:
	pusha                    ; must match 'macro exception'
	in al, PA_DAT
	test al, # OFFBUTTON     ; #on/off
	jnz offia                ; off, wait silently

	mov ax, # 2              ; #on, exception 2 aka NMI
	jmp near crash           ; or maybe jmp exceptions. pusha, mov ax, #

offia:
	mov al, # 0
	out PA_DAT, al           ; DISP_ON off

_reboot:
	xor ax, ax
	mov ds, ax               ; segment 0
	mov ss, ax
	mov sp, ax               ; avoid nmi cascade

	mov [v_nmi+0], # _reboot
	mov [v_nmi+2], cs
L9:
	cli
	hlt                  ; force hardware reset from watchdog.
	jmp L9

;----------------------------------------------------------------------

	macro interrupt_epilog   ; eoi_cmd
	mov ax, # ?1             ; EOI_xxx
	jmp near return_eoi
	mend

return_eoi:                  ; from interrupts
	mov dx, # EOI
	cli
	out dx, PCB_AX
	cmp sp, # intstack
	jb return                ; this was nested, quick out

	testb [astflag], # 3     ; anything posted ?
	jz no_ast
do_ast:
	sti

	testb [astflag], # 1     ; XXX could use D0 and D1 in PCB_IRR for these...
	jz no_tick
	andb [astflag], # ~1
	call _tick
no_tick:
	testb [astflag], # 2
	jz no_net
	andb [astflag], # ~2
	call _netintr
no_net:

	cli
	testb [astflag], # 3
	jnz do_ast               ; loop while new ones
no_ast:
	mov cx, [intstack+2]     ; system ?
	mov sp, [intstack+0]     ; XXX not nmiproof
	jcxz return              ; return to system level, ss stays 0

	mov bx, [_task]
	mov [bx+tsk_ss], cx      ; save task stack
	mov [bx+tsk_sp], sp
	mov sp, bx               ; empty system stack
	sti

return_task:                 ; from exceptions or syscall
	cmpb [runrun], # 0
	je no_switch
	call _swtch
no_switch:
	mov bx, [_task]
	cli
	mov ss, [bx+tsk_ss]
	mov sp, [bx+tsk_sp]
return:
	pop es
	pop ds
	popa
	iret

;----------------------------------------------------------------------

	macro interrupt_prolog   ; no arguments
	pusha                    ; must match IREGS definition
	push ds
	push es
	xor bp, bp               ; segment 0
	mov ds, bp
	mov bx, # intstack
	mov ax, ss
	cmp sp, bx               ; carry if sp is below intstack
	sbb ax, bp               ; carry if ss is also 0
	jc nested_?0             ; is a nested interrupt
	mov [bx+2], ss
	mov [bx+0], sp     ; task not updated yet (might be system or task)
	mov ss, bp
	mov sp, bx               ; bp:bx, 0:intstack
nested_?0:
	sti
	cld
	mov es, bp               ; all segregs 0
	mend


syscall:
	pusha                    ; XXX sub sp, # 10; push bp si di --- worth it ?
	push ds                  ; must match IREGS definition
	push es
	xor bp, bp               ; segment 0
	mov ds, bp
	mov bx, [_task]
	mov [bx+tsk_ss], ss      ; record task stack
	mov [bx+tsk_sp], sp      ; XXX crash if SS 0 XXX allow task_0...
	mov ss, bp
	mov sp, bx               ; empty system stack
	sti
	cld
	mov es, bp               ; all segregs 0
	mov si, # syscall_last
	cmp si, ax               ; syscall number
	jb badsys                ; syscalls[last] is _badsys()
	mov si, ax               ; in range
badsys:
	lds bx, [bx+tsk_sp]      ; doh... XXX
	push [bx+IREGS_SZ+8]     ; lift 4 argument words
	push [bx+IREGS_SZ+6]     ; enough for foo(int, long, int)
	push [bx+IREGS_SZ+4]
	mov ax, [bx+IREGS_SZ+2]  ; bx+...+0 is the near ret in task
	mov ds, bp               ; segment 0 again
	shl si, 1
	call _syscalls[si]       ; 
	mov bx, [_task]
	mov sp, bx               ; purge arguments, empty system stack
	lds bx, [bx+tsk_sp]
	mov [bx+IREGS_CX], # 0   ; XXX errno, later
	mov [bx+IREGS_DX], dx    ; dx is junk, if int return
	mov [bx+IREGS_AX], ax    ; return value
	xor bp, bp               ; segment 0
	mov ds, bp               ; segment 0 again
birth:                       ; forked child, empty system stack
justret:                     ; sigreturn etc
	jmp near return_task

exceptions:                  ; after pusha. number in ax
	mov cx, ss               ; SS already 0 ?
	jcxz crash               ; that's bad
	push ds                  ; must match IREGS definition
	push es
	xor bp, bp               ; segment 0
	mov ds, bp
	mov bx, [_task]
	mov [bx+tsk_ss], ss      ; record task stack
	mov [bx+tsk_sp], sp
	mov ss, bp
	mov sp, bx               ; empty system stack
	sti
	cld
	mov es, bp               ; all segregs 0
	push # return_task
	jmp near _exception      ; exception(number)

crash:
	push ds                  ; must match IREGS definition
	push es

	mov dx, sp
	push dx                  ; &frame
	push ss                  ; ss

	xor bp, bp               ; segment 0
	mov ss, bp
	mov sp, bp               ; empty safe stack
	mov ds, bp
	mov es, bp               ; all segregs 0
	cld
	call _dump               ; dump(number, ss, &frame)
	call _dump_asic
	jmp near _reboot

	macro exception          ; exception number
	pusha
	mov ax, # ?1
	jmp exceptions
	mend

;----------------------------------------------------------------------

divide_error:
	exception 0

single_step:
	exception 1

; NMI would be type 2

breakpoint:
	exception 3

overflow:
	exception 4

bounds_check:
	exception 5

invalid_opcode:
	exception 6

floating_escape:
	exception 7

; following numbers do not match intel 'types'

numerics:
	exception 8

reserved:
	exception 9

unknown:
	exception 10

;----------------------------------------------------------------------
	export _swtch

_swtch:
	mov bx, [_task]
	pushf
	pop [bx+sys_fl]
	pop [bx+sys_ip]          ; remove return address
	mov [bx+sys_sp], sp
	mov [bx+sys_bp], bp

_swtch_no_old:               ; from exit()
	cli
	movb [runrun], # 0

	mov bx, [_ready]
	mov bx, [bx+tsk_next]
	mov [_ready], bx         ; sigh...
	mov [_task], bx

	; XXX PA.BANK ?

	mov bp, [bx+sys_bp]
	mov sp, [bx+sys_sp]
	push [bx+sys_fl]
	push cs
	push [bx+sys_ip]         ; return address
	iret                     ; pops ip cs flags

;----------------------------------------------------------------------

	macro interrupt          ; eoi_cmd and handler
	interrupt_prolog
	import ?2
	call ?2
	interrupt_epilog ?1
	mend

int_3:
	interrupt EOI_INT3, _int_3

dma_0_int:
	interrupt EOI_DMA0, _dma_0_int

dma_1_int:
	interrupt EOI_DMA1, _dma_1_int

timer_0_int:
	interrupt EOI_TMR, _timer_0_int

timer_1_int:
	interrupt EOI_TMR, _timer_1_int

;----------------------------------------------------------------------
; 5 uarts

	import uart_0_int
	import uart_1_int
	import uart_2_int
	import uart_3_int
	import uart_4_int

int_0:
	interrupt_prolog

	in al, INT_0_REASON
again_0:
	mov bp, ax              ; XXX

	ror bp, # 1
	jnc not_0_0

	call uart_0_int
not_0_0:
	ror bp, # 1
	jnc not_0_1

	call uart_1_int
not_0_1:
	ror bp, # 1
	jnc not_0_2

	call uart_2_int
not_0_2:
	ror bp, # 1
	jnc not_0_3

	call uart_3_int
not_0_3:
	ror bp, # 1
	jnc not_0_4

	call uart_4_int
not_0_4:

	in al, INT_0_REASON
	or al, al
	jnz again_0

	interrupt_epilog EOI_INT0

;----------------------------------------------------------------------
; keyboard, keypad, barcode reader

	import _keyboard_rint
	import _keyboard_tint

int_1:
	interrupt_prolog

	in al, ASIC+0x1D             ; XXX status/interrupt enable reg
again_1:
	test al, # 2                 ; rint ?
	jz not_1_2

	push ax

	in al, ASIC+0x1A             ; data register
	call _keyboard_rint
	pop ax
not_1_2:
	test al, # 1                 ; XXX tint bit
	jz not_1_1

	call _keyboard_tint
not_1_1:
	in al, ASIC+0x1D             ; XXX status/interrupt enable reg
	test al, # 3                 ; r or tint ?
	jnz again_1

	interrupt_epilog EOI_INT1

;----------------------------------------------------------------------
; input pin and MBUS monitor, asic timer

	import _pa_chg
	import _pb_chg
	import _asic_timer_int
	import _netfree

int_2:
	interrupt_prolog

	in al, INT_2_REASON
again_2:
	mov bp, ax            ; XXX

	ror bp, # 1
	jnc not_2_0

	in ax, PA_DAT         ; PA_CHG to ah
still_2_0:
	call _pa_chg

	in ax, PA_DAT
	or ah, ah
	jne still_2_0         ; XXX no cure for the lockups. probably unnecessary

not_2_0:
	ror bp, # 1
	jnc not_2_1

	in ax, PB_DAT         ; PB_CHG to ah
still_2_1:
	call _pb_chg

	in ax, PB_DAT
	or ah, ah
	jne still_2_1         ; another one immediately

not_2_1:
	ror bp, # 1
	jnc not_2_2

	in al, ASIC+0x29      ; asic timer
	call _asic_timer_int  ; XXX similar cures needed here ?
not_2_2:
	ror bp, # 1
	jnc not_2_3

	in al, ASIC+0x18      ; MBUS (and ?) idle monitor
	call _netfree
not_2_3:

	in al, INT_2_REASON
	or al, al
	jnz again_2

	interrupt_epilog EOI_INT2

;----------------------------------------------------------------------
; HZ periodic

timer_2_int:
	interrupt_prolog

	mov al, # 1
	out ASIC+0x2A, al     ; holds off watchdog in ASIC by 1 second

	dec [_fido]
	jz reboot_            ; 10 seconds since idle() last ran

	mov bx, # _timex
	add [bx+0], # (0x10000 + HZ/2) / HZ
	adc [bx+2], # 0
	adc [bx+4], # 0

	orb [astflag], # 1   ; tick
	movb [runrun], # 1   ; switch

	interrupt_epilog EOI_TMR

reboot_:
	jmp near _reboot

;----------------------------------------------------------------------
	export _disable, _restore, _enable
	export _spl_up, _splx
	export _inb, _outb
	export _inw, _outw_pcb

_disable:             ; word disable();
	pushf
	pop ax            ; oflags
	cli
	ret

_restore:             ; void restore(word oflags);
	test ah, # 0x02   ; ARG1. FLAGS.IF ?
	jz L4
_enable:              ; void enable();
	sti
L4:
	ret

_splx:                ; void splx(opri);
	mov dx, # IPRIO
	out dx, PCB_AX    ; ARG1
	ret

_spl_up:              ; word spl_up(npri);
	mov bx, ax        ; ARG1
	mov dx, # IPRIO
	in ax, dx         ; opri
	cmp ax, bx
	jbe L3
	xchg ax, bx
	out dx, PCB_AX
	xchg ax, bx
L3:
	ret

_inb:                 ; word inb(port);
	mov dx, ax        ; ARG1
	mov ah, # 0       ; zeroextend
	in al, dx
	ret

_inw:                 ; word inw(port);
	mov dx, ax        ; ARG1
	in ax, dx
	ret

_outb:                ; void outb(port, byte);
	mov dx, ax        ; ARG1
	ARG2 al
	out dx, al
	ret

_outw_pcb:            ; void outw_pcb(port, word);
	mov dx, ax        ; ARG1
	ARG2 ax
	out dx, PCB_AX
	ret

;----------------------------------------------------------------------

_dma:               ; void dma(int ch, word con, word cc, dword src, dword dst)
	push bp         ; 0  2  4   6  8    12
	mov bp, sp      ; bp ra con cc s:rc d:st

	or  ax, ax
	mov dx, # D0SRCL
	jz  dma_0
	mov dx, # D1SRCL
dma_0:
	mov ax, [bp+8]
	out dx, PCB_AX  ; DxSRCL
	inc dx
	inc dx
	mov ax, [bp+10]
	out dx, PCB_AX  ; DxSRCH
	inc dx
	inc dx
	mov ax, [bp+12]
	out dx, PCB_AX  ; DxDSTL
	inc dx
	inc dx
	mov ax, [bp+14]
	out dx, PCB_AX  ; DxDSTH
	inc dx
	inc dx
	mov ax, [bp+6]
	out dx, PCB_AX  ; DxCNT
	inc dx
	inc dx
	mov ax, [bp+4]
	out dx, PCB_AX  ; DxCON

	leave
	ret

;----------------------------------------------------------------------
; for delays shorter than 16 msec

	export _busywait_usec

_busywait_usec:
	mov bx, # 256 * TMRCLK / 1000000 ; usec to ticks of Timer 1
	mul bx
	mov bl, ah                       ; /= 256
	mov bh, dl
	mov dx, # T1CNT
	in  ax, dx
	add bx, ax                       ; target ticks
not_yet:
	in  ax, dx
	cmp bx, ax
	jns not_yet
	ret

;----------------------------------------------------------------------

_sys_getdtablesize:
	mov ax, # NOFILE
	ret

_sys_getpid:
	mov ax, [_task]
	ret

_sys_time:
	mov cx, es
	les ax, [_timew]  ; atomic dword
	mov dx, es
	mov es, cx
	add ax, [_boottime+2]
	adc dx, [_boottime+4]
	ret

_sys_gettimeofday:       ; gettimeofday(tvp);
	mov di, ax           ; tvp

	mov bx, # _timex
	cli
	mov cx, [bx+0]       ; cx 1/65536 seconds
	mov ax, [bx+2]       ; dxax seconds
	mov dx, [bx+4]
	sti

	mov bx, # _boottime
	add cx, [bx+0]
	adc ax, [bx+2]
	adc dx, [bx+4]

	mov bx, [_task]
	push ds
	mov ds, [bx+tsk_ss]  ; sulong
	mov [di+0], ax
	mov [di+2], dx       ; tv_sec

	mov ax, # 1000000 * 256 / 0x10000
	mul cx
	mov [di+4], ah
	mov [di+5], dl
	mov [di+6], dh
	mov [di+7], # 0      ; tv_usec
	pop ds
	ret

_sys_settimeofday:       ; settimeofday(tvp);
	mov di, ax           ; tvp

	mov bx, [_task]
	push ds
	mov ds, [bx+tsk_ss]  ; fulong
	mov ax, [di+4]
	mov dx, [di+6]       ; tv_usec dxax

	mov cx, # 4294     ; XXX argh sigh 0x10000 * 0x10000 / 1000000
	mul cx             ; 1/65536 seconds in dx

	mov ax, [di+0]
	mov cx, [di+2]       ; tv_sec cxax (NOT dxax)
	pop ds

	mov bx, # _timex
	cli
	sub dx, [bx+0]
	sbb ax, [bx+2]
	sbb cx, [bx+4]
	sti

	mov bx, # _boottime
	mov [bx+0], dx
	mov [bx+2], ax
	mov [bx+4], cx
	ret

_sys_yield:
	call _swtch
	ret

;----------------------------------------------------------------------
	export _sys_exit

_sys_exit:                  ; no args, no return
	mov bx, [_task]
	call remrun

	; running on old task stack for a short while

	mov dx, [free]          ; XXX task_relse:
	mov [bx+tsk_next], dx
	mov [free], bx

	; XXX release task data segment

	jmp near _swtch_no_old

;----------------------------------------------------------------------
; arguments of handler are in stack.
; task must fix things by itself,
; if it uses FIRST_ARG_IN_AX etc

_sendsig:                  ; void sendsig(handler, sig, code)
	ARG3 cx, dx            ; sig and code
	mov bx, [_task]
	push ds
	lds si, [bx+tsk_sp]    ; current frame
	mov di, si             ; tack more stuff below it

	; sigcontext, handler and arguments

	sub di, # 6
	mov [di+4], # 0               ; XXX hmm?
	mov [di+2], # 0               ; XXX omask
	mov [di+0], si                ; return frame
	mov [di-2], di                ; void *scp
	mov [di-4], dx                ; int code;
	mov [di-6], cx                ; int sig;
	mov [di-8], ax                ; handler
	sub di, # 8

	; frame to return to, now

	sub di, # IREGS_SZ
	mov [di+IREGS_FL], # 0x200            ; FLAGS.IF
	mov [di+IREGS_IP], # task_sigtramp    ; XXX in task cs *scratch*
	mov [di+IREGS_BP], # 0                ; XXX maybe bp from old frame ?

	mov ax, [si+IREGS_CS]      ; segregs.
	mov [di+IREGS_CS], ax
	mov [di+IREGS_ES], ds
	mov [di+IREGS_DS], ds      ; other regs junk (dxax & cx set by syscall ret)

	pop ds
	mov [bx+tsk_sp], di
	ret

_sys_sigreturn:             ; sigreturn(scp)
	mov di, ax
	mov bx, [_task]
	push ds
	mov ds, [bx+tsk_ss]     ; fuword
	mov si, [di+0]          ; return frame
	mov ax, [di+2]          ; XXX omask
	mov dx, [di+4]          ; XXX hmm?

	pop ds
	mov [bx+tsk_sp], si     ; rewind to the return frame

	pop ax                  ; skip a bit,
	jmp near justret        ; no return value, just return

; XXX template, should be in task cs... *scratch*

task_sigtramp:              ; stack: handler, sig, code, scp
	pop bx                  ; handler
	call bx                 ; handler(sig, code, scp);
	pop ax                  ; purge 'int sig',
	mov ax, # SYS_sigreturn ; but leave 'int code' on stack,
	int 0x20                ; to act as the near return. sigreturn(scp).
	jmp 0+*                 ; jmp .
task_sigtramp_end:

;----------------------------------------------------------------------
; XXX 'spork', fork without separate dseg, to given sp

ffail:
	mov ax, # -1
	ret

_sys_fork:
	mov di, [free]          ; XXX task_alloc:
	or di, di
	jz L5                   ; must allocate new task structure
	mov ax, [di+tsk_next]
	mov [free], ax
	jmp L6
L5:
	cmp [_farbrk], # 0x6000 ; XXX
	jae ffail
	mov ax, [_brk]
	add ax, # SYS_STK
	mov di, ax
	add ax, # tsk_size
	mov [_brk], ax

	mov ax, [_farbrk]       ; XXX XXX
	mov [di+tsk_ss], ax
	add ax, # 0x1000
	mov [_farbrk], ax
L6:

	mov [di+sys_ip], # birth  ; where child resumes
	mov [di+sys_fl], # 0x200  ; interrupts enabled
	mov [di+sys_bp], # 0
	mov [di+sys_sp], di       ; empty system stack

	mov si, [_task]           ; parent
	mov bx, [si+tsk_sp]       ; saved task registers
	mov [di+tsk_sp], bx       ; same offset, different segment

	mov cx, [si+tsk_bound+0]
	mov ax, [si+tsk_bound+2]
	mov [di+tsk_bound+0], cx
	mov [di+tsk_bound+2], ax

	mov cx, [si+tsk_dsize+0]  ; bytes
	mov ax, [si+tsk_dsize+2]
	mov [di+tsk_dsize+0], cx
	mov [di+tsk_dsize+2], ax


	push ds
	push es
	push si
	push di
	mov es, [di+tsk_ss]
	mov ds, [si+tsk_ss]

	; XXX PA.BANK XXX could be different between parent and child...
	; could copy from sp wrapping over 0 to 'break'
	; but this would require a fixed 64kB dseg.
	; and 'break' is not known here

	xor si, si
	xor di, di
	shr ax, 1               ; words XXX loop if dsize over 64 kB
	rcr cx, 1
	rep                     ; uuh...
	movsw                   ; copy dataseg

	mov ax, es              ; patch child
	mov ds, ax
	mov [bx+IREGS_DS], ds   ; DS=ES=SS
	mov [bx+IREGS_ES], ds   ; child segregs

	mov [bx+IREGS_CX], # 0  ; XXX errno, later
	mov [bx+IREGS_DX], # 0  ; child return value, 0
	mov [bx+IREGS_AX], # 0

	pop di
	pop si
	pop es
	pop ds

	mov bx, di
	call insrun

	mov ax, di              ; return task struct as a 'pid'
	ret

;----------------------------------------------------------------------

rwtab:
	.word _getc_0,   _putc_0
	.word _getc_1,   _putc_1
	.word _getc_2,   _putc_2
	.word _getc_3,   _putc_3
	.word _getc_4,   _putc_4
	.word _getc_kbd, _putc_screen

	rwtab_count equ (* - rwtab) / 4

rwfail:
	mov ax, # -1
	leave
	ret

_sys_write:
	push bp
	mov bp, sp
	cmp ax, # rwtab_count
	jae rwfail
	mov bx, ax
	shl bx, 2
	push 2+rwtab[bx]
	push [bp+6]
	jmp wr2
wr1:
	mov si, [bp+4]
	call fubyte
	mov [bp+4], si
	call [bp-2]
wr2:
	dec [bp+6]
	jns wr1

	cmp [bp-2], # _putc_screen
	jne wr3
	call _update_cursor
wr3:
	mov ax, [bp-4]
	leave
	ret

_sys_read:
	push bp
	mov bp, sp
	cmp ax, # rwtab_count
	jae rwfail
	mov bx, ax
	shl bx, 2
	push 0+rwtab[bx]
	push # 0
	jmp rd2
rd1:
	call [bp-2]
	test ax, ax
	js rd3
	mov di, [bp+4]
	call subyte
	mov [bp+4], di
	inc [bp-4]
rd2:
	dec [bp+6]
	jns rd1
rd3:
	mov ax, [bp-4]
	leave
	ret

;----------------------------------------------------------------------

_sys_sleep:
	ret

;----------------------------------------------------------------------

	export _sleep
	export _wakeup

_sleep:                      ; sleep(&queue)
	mov si, ax

sleep:
	mov bx, [_task]
	call remrun

	mov dx, [si]
	mov [bx+tsk_next], dx
	mov [si], bx

	call _swtch              ; return later
	ret


_wakeup:                     ; wakeup(&queue)
	mov si, ax
	jmp wakeup

wup1:
	movb [runrun], # 1       ; switch

	mov ax, [bx+tsk_next]
	mov [si], ax
	call insrun
wakeup:
	mov bx, [si]
	test bx, bx
	jnz wup1                 ; a sleeper
	ret

;----------------------------------------------------------------------

insrun:                      ; bx
	push si
	push di
	pushf
	cli
	mov si, [_ready]         ; 1st
	mov di, [si+tsk_prev]    ; last
	mov [bx+tsk_next], si
	mov [bx+tsk_prev], di
	mov [si+tsk_prev], bx
	mov [di+tsk_next], bx
	popf
	pop di
	pop si
	ret

remrun:                      ; bx
	push si
	push di
	pushf
	cli
	mov si, [bx+tsk_prev]
	mov di, [bx+tsk_next]
	mov [si+tsk_next], di
	mov [di+tsk_prev], si
	cmp bx, [_ready]
	jne rr1
	mov [_ready], di
rr1:
	popf
	pop di
	pop si
	ret

;----------------------------------------------------------------------
	export _fubyte
	export _fuword
	export _fulong
	export _subyte
	export _suword
	export _sulong
	export _copyin
	export _copyout

_fubyte:                         ; between here ...
	mov si, ax
fubyte:
	mov bx, [_task]
;	cmp si, [bx+tsk_bound+XXX]     XXX
;	ja fault_ret
	mov dx, ds
	mov ds, [bx+tsk_ss]
	lodsb
	mov ds, dx
	mov ah, # 0
	ret

_fuword:
	mov si, ax
fuword:
	mov bx, [_task]
;	cmp si, [bx+tsk_bound+XXX]     XXX
;	jae fault_ret
	mov dx, ds
	mov ds, [bx+tsk_ss]
	lodsw
	mov ds, dx
	ret

_fulong:
	mov si, ax
fulong:
	mov bx, [_task]
;	cmp si, [bx+tsk_bound+XXX]     XXX
;	jae fault_ret
	mov cx, ds
	mov ds, [bx+tsk_ss]
	lodsw
	mov dx, [si]
	add si, # 2
	mov ds, cx
	ret

_subyte:
	mov di, ax
	ARG2 al
subyte:
	mov bx, [_task]
;	bound di, [bx+tsk_bound]     XXX
	mov dx, es
	mov es, [bx+tsk_ss]
	stosb
	mov es, dx
	xor ax, ax
	ret

_suword:
	mov di, ax
	ARG2 ax
suword:
	mov bx, [_task]
;	bound di, [bx+tsk_bound]     XXX
	mov dx, es
	mov es, [bx+tsk_ss]
	stosw
	mov es, dx
	xor ax, ax
	ret

_sulong:
	mov di, ax
	ARG3 ax, dx
sulong:
	mov bx, [_task]
;	bound di, [bx+tsk_bound]     XXX
	mov cx, es
	mov es, [bx+tsk_ss]
	stosw
	eseg
	mov [di], dx
	add di, # 2
	mov es, cx
	xor ax, ax
	ret

_copyin:
	mov si, ax
	ARG3 di, cx
copyin:
	mov bx, [_task]
;	mov ax, si
;	add ax, cx
;	jc fault_ret
;	cmp ax, [bx+tsk_bound+XXX]     XXX
;	ja fault_ret
	mov dx, ds
	mov ds, [bx+tsk_ss]
	shr cx, 1
	rep
	movsw
	rcl cx, 1
	rep
	movsb
	mov ds, dx
	xor ax, ax
	ret

_copyout:
	mov si, ax
	ARG3 di, cx
copyout:
	mov bx, [_task]
;	cmp di, [bx+tsk_bound+YYY]     XXX
;	jb fault_ret
;	mov ax, di
;	add ax, cx
;	jc fault_ret
;	cmp ax, [bx+tsk_bound+XXX]     XXX
;	ja fault_ret
	mov dx, es
	mov es, [bx+tsk_ss]
	shr cx, 1
	rep
	movsw
	rcl cx, 1
	rep
	movsb
	mov es, dx
	xor ax, ax
	ret

fault_ret:                      ; ... and here, bounds check continues here
	or ax, # -1
	cwd
	ret                         ; nz

; early in bound exception
	mov bp, sp
	mov cx, cs
	cmp cx, [bp+IREGS_CS]
	jne not_fusu
	mov cx, [bp+IREGS_IP]
	sub cx, # _fubyte
	cmp cx, # fault_ret - _fubyte
	jae not_fusu
	mov [bp+IREGS_IP], # fault_ret
	jmp near return
not_fusu:

;----------------------------------------------------------------------
; two 256 kB banks, each divided to 16 pieces.
; 16384 bytes, or 1024 paragraphs, per piece.

	export _alloc_seg
;	export _free_seg

_alloc_seg:               ; seg alloc_seg(paragraphs, bank)
	ARG2 bx               ; _bank_1 or _2
	add ax, # 1023
	shr ax, 10            ; (paragraphs + 1023) / 1024
	mov si, ax
	shl si, 1
	mov dx, 2+_wordmask[si]  ; 0x0002 if 1 piece
	dec dx                   ; allocation as a bitmask

	mov ax, # 0x2000      ; base segment of both banks
as_2:
	test [bx], dx
	jz as_3
	add ax, # 1024        ; one piece in paragraphs
	shl dx, 1
	jnc as_2
	xor ax, ax            ; return 0, no space
	ret
as_3:
	or [bx], dx           ; mark as used
	ret                   ; return base paragraph

;----------------------------------------------------------------------
; XXX interbank copyseg

	IF 0

	mov al, [si+tsk_bank]
	or  al, [di+tsk_bank]
	cmp al, # 3
	...


copyseg_1:
	lds si, [bp-4]
	les di, [bp-12]
	mov cx, # 256 / 2
	rep
	movsw

	cli
	in al, PA_DAT
	xor al, # BANK
	out PA_DAT, al
	sti

	lds si, [bp-12]
	les di, [bp-8]
	mov cx, # 256 / 2
	rep
	movsw

	cli
	in al, PA_DAT
	xor al, # BANK
	out PA_DAT, al
	sti

	add [bp-4], # 256
	add [bp-8], # 256
	dec [bp-14]
	jnz copyseg_1

	ENDIF

;----------------------------------------------------------------------

	.data
_syscalls:
	.word _sys_exit          ; 0 must match syscalls.h
	.word _sys_yield         ; 1
	.word _sys_fork          ; 2
	.word _sys_getpid        ; 3
	.word _sys_time          ; 4
	.word _sys_sleep         ; 5
	.word _sys_read          ; 6
	.word _sys_write         ; 7
	.word _sys_sigreturn     ; 8
	.word _sys_getdtablesize ; 9
	.word _sys_gettimeofday  ; 10
	.word _sys_settimeofday  ; 11
	.word _sys_badsys        ; must be last

syscall_last equ (* - _syscalls) / 2 - 1

