
/* KP3BC and AT59FI on D15 */

#include "head.h"
#include "pcb.h"
#include "asic.h"
#include "keys.h"

#define BUFSIZ 128

static byte kbd_buf[BUFSIZ];
static word
	kbd_ins,
	kbd_rem;

/* KP3BC */
static byte letters[] = "\
0oqz0\
1 .,?!-+=_/:;1\
2abc2\
3def3\
4ghi4\
5jkl5\
6mn6\
7prs7\
8tuv8\
9wxy9\
";

static byte *arrows[] = {
	"\033OA",
	"\033OB",
	"\033OD",
	"\033OC",
	"\033OH", /* home */
	"\033OF", /* end */
};

/* KLCK 11, KCTL E0 11. other E0 xx have unique xx */

static byte
keymap[] = {
	0,    '#',  0,    KF5,  KF3,  KF1,  KF2,  0,     /* 0x */
	0,    KSND, '*',  KF6,  KF4,  9,    '`',  0,

	0,    KLCK, KSFT, 0,    0,    'q',  '1',  0,     /* 1x */
	0,    0,    'z',  's',  'a',  'w',  '2',  0,

	0,    'c',  'x',  'd',  'e',  '4',  '3',  0,     /* 2x */
	0,    ' ',  'v',  'f',  't',  'r',  '5',  0,

	0,    'n',  'b',  'h',  'g',  'y',  '6',  0,     /* 3x */
	0,    0,    'm',  'j',  'u',  '7',  '8',  0,

	0,    ',',  'k',  'i',  'o',  '0',  '9',  0,     /* 4x */
	0,    '.',  '-',  'l',  '',  'p',  '+',  0,

	0,    0,    '',  0,    '',  '=',  0,    0,     /* 5x */
	0,    0,    '\r', ']',  0,    '\\', 0,    0,

	0,    0,    0,    0,    0,    0,    8,    0,     /* 6x */
	0,    KEND, 0,    KLFT, KHME, 0,    0,    0,

	'0',  '.',  KDN,  '5',  KRHT, KUP,  0x1B, 0,     /* 7x */
	0,    '+',  '3',  '-',  '*',  '9',  KEND, 0,

	0,    0,    0,    KF7
},
shifted[] = {
	0,    '#',  0,    KF5,  KF3,  KF1,  KF2,  0,     /* 0x */
	0,    KSND, '*',  KF6,  KF4,  9,    '~',  0,

	0,    KLCK, KSFT, 0,    0,    'Q',  '!',  0,     /* 1x */
	0,    0,    'Z',  'S',  'A',  'W',  '"',  0,

	0,    'C',  'X',  'D',  'E',  '$',  '#',  0,     /* 2x */
	0,    ' ',  'V',  'F',  'T',  'R',  '%',  0,

	0,    'N',  'B',  'H',  'G',  'Y',  '&',  0,     /* 3x */
	0,    0,    'M',  'J',  'U',  '/',  '(',  0,

	0,    ';',  'K',  'I',  'O',  '=',  ')',  0,     /* 4x */
	0,    ':',  '\'', 'L',  '',  'P',  '?',  0,

	0,    0,    '',  0,    '',  '=',  0,    0,     /* 5x */
	0,    0,    '\n', ']',  0,    '\\', 0,    0,

	0,    0,    0,    0,    0,    0,    127,  0,     /* 6x */
	0,    KEND, 0,    KLFT, KHME, 0,    0,    0,

	'0',  '.',  KDN,  '5',  KRHT, KUP,  0x1B, 0,     /* 7x */
	0,    '+',  '3',  '-',  '*',  '9',  KEND, 0,

	0,    0,    0,    KF7
},
control[] = {
	0,    '#',  0,    KF5,  KF3,  KF1,  KF2,  0,     /* 0x */
	0,    KSND, '*',  KF6,  KF4,  9,    0x07, 0,

	0,    KLCK, KSFT, 0,    0,    0x11, '~',  0,     /* 1x */
	0,    0,    0x1A, 0x13, 0x01, 0x17, '@',  0,

	0,    0x03, 0x18, 0x04, 0x05, '`',  '^',  0,     /* 2x */
	0,    0x00, 0x16, 0x06, 0x14, 0x12, '|',  0,

	0,    0x0E, 0x02, 0x08, 0x07, 0x19, '*',  0,     /* 3x */
	0,    0,    0x0D, 0x0A, 0x15, '{',  '[',  0,

	0,    0x1E, 0x0B, 0x09, 0x0F, '}',  ']',  0,     /* 4x */
	0,    0x1F, '_',  0x0C, 0x1D, 0x10, '\\', 0,

	0,    0,    0x1B, 0,    0x1C, '=',  0,    0,     /* 5x */
	0,    0,    '\n', ']',  0,    '\\', 0,    0,

	0,    0,    0,    0,    0,    0,    127,  0,     /* 6x */
	0,    KEND, 0,    '<',  KHME, 0,    0,    0,

	'0',  '.',  KDN,  '5',  '>',  KUP,  0x1B, 0,     /* 7x */
	0,    '+',  '3',  '-',  '*',  '9',  KEND, 0,

	0,    0,    0,    KF7
};

static void
arrow(x)
{
	byte *cp = arrows[x];

	while (x = *cp++)
		stash(x);
}

/* E0 ... FE are lowercase in 8859 */

#define _islower(c) (((c) >= 'a' && (c) <= 'z') || (c) >= 0xE0)

/* during input interrupt. shift state is not lost
 * if fifo overflows */

void
keycode(code)
	byte code;
{
	static byte flags;
#define F0  1
#define E0  2
#define LCK 4
#define SFT 8
#define CTL 0x10

	byte c;

#if 0
	printf(" %x ", code);
#endif

	if (code >= sizeof(keymap)) {

		/* failure would be FC,
		 * barcode reader emits FF, anything else ? */

		switch (code) {

		case 0xAA:       /* power on */
			flags = 0;
			break;

		case 0xE0:       /* extended */
		case 0xE1:
			flags |= E0;
			break;

		case 0xF0:       /* break prefix */
			flags |= F0;
			break;
		}
	} else {
		c = keymap[code];

		if (flags & E0) {
			if (c == KLCK)  /* KLCK 11, KCTL E0 11 */
				c = KCTL;
		}
		if (flags & F0) {
			switch (c) {
			case KCTL:
				flags &= ~CTL;
				break;
			case KSFT:
				flags &= ~SFT;
				break;
			case KLCK:
				flags ^= LCK;        /* lock break toggles the state */
				break;
			}
		} else {
			switch (c) {
			case KCTL:
				flags |= CTL;
				break;
			case KSFT:
				flags |= SFT;
			case KLCK:                /* lock make is ignored */
				break;

			default:
				if (flags & CTL)
					c = control[code];
				else
				if (flags & SFT)
					c = shifted[code];
				else
				if ((flags & LCK) && _islower(c))
					c ^= 0x20;
/* XXX */
#if 1
				if (c >= KUP && c <= KEND)
					arrow(c - KUP);
				else
#endif
				stash(c);
			}
		}
		flags &= ~(F0 | E0);
	}
}

void
keyboard_tint()
{
	printf(" kti ");
}

void
keyboard_rint(b)
	byte b;
{
	static byte nbits;
	static dword accu; /* must hold 10+8 bits */
	word frame;

	accu = (accu << 8) | b;
	nbits += 8;

	while (nbits >= 11) {
		nbits -= 11;
		frame = accu >> nbits;

		if (valid(frame)) {
			keycode(revbits[(frame >> 2) & 0xFF]);
			break;
		}
		nbits += 10; /* try to resync */
	}
}

#asm
	.text

_valid:
	test ah, # 4      ; startbit
	jnz vout          ;  not 0. carry cleared by TEST
	shr ax, # 1       ; stopbit
	jnc vout          ;  not 1
 	shr ax, # 1       ; paritybit in CF, keycode in AL, PF set from AL (not AX)
	jpe vout          ;  even keycode, paritybit must have been 1
	cmc               ;  odd keycode, paritybit now 1, if ok
vout:
	sbb ax, ax        ; carry was clear, if invalid
	ret

_stash:
	mov bx, [_kbd_ins]
	mov _kbd_buf[bx], al
	inc bx
	and bx, # BUFSIZ - 1
	mov [_kbd_ins], bx
	ret

;----------------------------------------------------------------------
	export _getc_kbd

_getc_kbd:
	xor ax, ax
	mov bx, [_kbd_rem]
	cmp bx, [_kbd_ins]
	jne got
	dec ax                  ; -1
	ret
got:
	mov al, _kbd_buf[bx]
	inc bx
	and bx, # BUFSIZ - 1
	mov [_kbd_rem], bx
	ret                     ; 0...255

;----------------------------------------------------------------------
; input is received in unaligned reversed chunks of 6 (six) bits.
; each keycode is startbit, 8 databits, odd parity and stopbit.

;	export _keyboard_rint

	.lcomm accu,  2
	.lcomm nbits, 1

XXX_keyboard_rint:
	and al, # 0x3F       ; 6 new bits
	mov cl, # 6
	mov bx, [accu]       ; 10 or less old bits
	shl bx, cl
	or bl, al
	mov [accu], bx
	add cl, [nbits]      ; now this many bits
again:
	mov [nbits], cl
	sub cl, # 11
	jc not_enough
	mov ax, bx           ; align the full frame to ax
	shr ax, cl
	test ah, # 4         ; startbit
	jnz resync           ;  not 0
	shr ax, # 1          ; stopbit
	jnc resync           ;  not 1
	shr ax, # 1          ; databits to AL, paritybit to CF, parity to PF
	lahf
	test ah, # 5         ; PF and CF
	jpo resync           ;  do not match
	mov [nbits], cl      ; left for next time

	mov bx, # _revbits
	xlatb
	jmp near _keycode
not_enough:
	ret

resync:
	add cl, # 10         ; drop oldest bit
	jmp again

