
; XXX there is time to do 15600 Hz

	.module afsk

	.include "iomap.inc"
	.include "cfg.inc"

; 14400 Hz equals
; 280 clocks of TMR_MT,
; 280 T on P8N and
; 560 T on P8E

PWM = 14400

TONE1 = 1200 * 0x100 / PWM
TONE2 = 2200 * 0x100 / PWM ; XXX check by hand if round up is better

	.area _BSS

old_gain: .ds 1

	.area _CODE

;----------------------------------------------------------------------

	.IF 0

	ld hl, #-100
	add hl, sp
	ld sp, hl           ; pkt

	ld de, "QST"
	call axaddr
	ld de, # _cfg + mycall
	call axaddr

	dec hl
	set 0, (hl)         ; Last
	inc hl
	ld (hl), # 0x03     ; UI Text
	inc hl
	ld (hl), # 0xF0
	inc hl

	ex de, hl
	ld bc               ; datalen
	ld hl               ; data
	ldir

	ld hl, # 0
	add hl, sp          ; pkt
	ex de, hl           ;  and end
	and a
	sbc hl, de          ; pkt and len

	push hl             ; len
	push de             ; pkt
	call _ax_send       ;

	ld hl, # 100 + 4
	add hl, sp
	ld sp, hl
	ret

	.ENDIF

_ax_send::                  ; void ax_send(byte *pkt, word len);

	ld hl, # 1200           ; abort tone, just in case
	call timer_mt
	call _CCIRC_on

	ld a, (_cfg + afsk_gain)
	ld hl, # old_gain
	cp (hl)
	jr z, 1$                ; same as old
	ld (hl), a

	; pwm slower than 15750 (256 timer clocks),
	; full range in sine is ok

	ld e, a                 ; gain
	ld d, # 0x80            ; center
	ld iy, # _afsk_sine     ; tab
	call make_sine
1$:
	ld hl, (_cfg + ax_txdelay) ; msec
	srl h                      ; to doublebits, approx
	rr l
	ld e, l
	ld d, h
	srl d
	rr e
	srl d
	rr e
	add hl, de              ; idle = txdelay / 2 + txdelay / 8

	pop af                  ; arguments:
	pop de                  ; pkt
	pop bc                  ; len
	push bc
	push de
	push af

	push bc                 ; len without crc
	push de                 ; pkt
	push hl                 ; idle

	call crc_ccitt          ; HL = ~crc(DE, BC) and DE = end
	ex de, hl               ; HL end, DE ~crc
	ld (hl), e
	inc hl
	ld (hl), d              ; pkt must have room for crc word XXX

	pop bc                  ; idle
	pop hl                  ; pkt
	pop de                  ; len without crc
	inc de
	inc de                  ;  with
	call axout              ; HL pkt, DE len, BC idle

	call _CCIRC_off
	jp _timer_mt_off

;----------------------------------------------------------------------

_axout::                ; void axout(byte *, word len, word txdelay);
	pop af
	pop hl              ; HL packet
	pop de              ;    length
	pop bc              ; BC idlebits / 2
	push bc
	push de
	push hl
	push af
axout:

	rst DISABLE         ; from now on.

	ex af, af'
	exx
	push af             ; save altset
	push hl
	push bc
	push de

	ld d, # TONE1       ; D' 1200
	ld h, # TONE2       ; H' 2200
	ld a, # 0           ; A' phase acc
	ld b, #>_afsk_sine  ; B' hi(sine)
	exx
	ex af, af'

	call doit

	pop de              ; restore altset
	pop bc
	pop hl
	pop af
	exx
	ex af, af'

	jp _restore         ; RET


doit:
	push de             ; length in stack until idle and flag sent

	; E downcount to stuffbit
	; A D C' E' L' scratch regs

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	; use similar opcodes for dummies to
	; avoid differences between P8E and N

	inc bc              ;      0 is short delay, 65535 long
1$:
	call fsk_0          ;    X zerobit
	inc de              ;  6   dummy (E can change, next is zerobit)
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_0          ;    X zerobit, start of flag maybe
	dec bc              ;  6 X --padbits
	ld a, c             ;  4 X
	or b                ;  4 X
	jp nz, 1$           ; 10 X while padbits

	call fsk_1          ;    X rest of starting flag
	inc de              ;  6   dummy (also, undo stuffbit downcount)
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_0          ;    X

	inc bc              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	pop bc              ; 10 X BC packet length
2$:
	call fsk_bit        ;    X databits
	inc bc              ;  6   dummy ++length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	dec bc              ;  6   dummy --length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	inc bc              ;  6   dummy ++length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	dec bc              ;  6   dummy --length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	inc bc              ;  6   dummy ++length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	dec bc              ;  6   dummy --length
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	dec bc              ;  6 X  --length for real
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_bit        ;    X
	inc hl              ;  6 X ++ptr (avoiding CPI, prefixed)
	ld a, c             ;  4 X
	or b                ;  4 X
	jp nz, 2$           ; 10 X while length

	call fsk_0          ;    X end flag
	inc de              ;  6   dummy (also, undo stuffbit downcount)
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X
	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_0          ;    X

	inc de              ;  6   dummy
	nop                 ;  4   dummy
	nop                 ;  4   dummy
	jp .+3              ; 10   dummy
	call fsk_1          ;    X fade out

	ret                 ; done


fsk_1:
	ld a, # 0           ;  7 dummy
	nop                 ;  4 dummy
	ld a, # 0           ;  7 dummy
	jp fsk_1_           ; 10

fsk_0:
	ld a, # 0           ;  7 dummy
	nop                 ;  4 dummy
	ld a, # 0           ;  7 dummy
	jp fsk_0_           ; 10

fsk_bit:
	ld a, (hl)          ;  7 bytes rotate back to original
	rrca                ;  4 value after a byte is sent
	ld (hl), a          ;  7 rrc (hl) would be awkward with P8E
	jp c, fsk_1_        ; 10

fsk_0_:
	ld e, # 5 + 1       ;  7 so many onebits to stuffbit
	exx                 ;  4
	ex de, hl           ;  4 swap H' D' phase increments
	jp bit              ; 10

fsk_1_:
	ld a, # 0           ;  7 dummy
	exx                 ;  4
	nop                 ;  4 dummy
	jp bit              ; 10

bit:
	ld e, # PWM / 1200  ;  7 so many slots per bittime
slot:
	; waste cycles

	ld a, (_logic)     ;; 13/14
	ld a, (_logic)     ;; 13/14
	ld a, (_logic)     ;; 13/14
	ld a, (_logic)     ;; 13/14
	jp .+3             ;; 10/11

	ld a, (_logic)     ;; 13/14
	cp # 0             ;;  7/8
	call z, burn_more  ;; 10/18

	ex af, af'         ;;  4/5
	add a, d           ;;  4/5 acc += inc
	ld c, a            ;;  4/5
	ex af, af'         ;;  4/5

	ld a, (bc)         ;;  7/8 afsk_sine[]
	out (TMR_MT), a    ;; 11/12
	out (WDOG), a      ;; 11/12

	dec e              ;;  4/5
	jp nz, slot_       ;; 10/11 do the bittime
	exx
	dec e
	jp z, stuffbit
	ret

stuffbit:               ; more dummies. a legion.
	inc de              ;  6 axout: E is zero and reloaded in fsk_0
	nop                 ;  4
	nop                 ;  4
	jp .+3              ; 10
	call dummy          ; 17
	jp fsk_0            ; 10

slot_:
	inc de             ;;  6/7 axout: E' value something small here
	dec e              ;;  4/5        and back
	nop                ;;  4/5
	jp .+3             ;; 10/11
	call dummy         ;; 17/18 call (& ret)
	ld a, # 0          ;;  7/8 fsk_bit:
	nop                ;;  4/5
	ld a, # 0          ;;  7/8
	jp .+3             ;; 10/11
	ld a, # 0          ;;  7/8 fsk_bit_:
	nop                ;;  4/5
	nop                ;;  4/5
	jp .+3             ;; 10/11
	ld a, # 0          ;;  7/8 bit: ld e, # slotcount
	nop                ;;  4/5 exx
	nop                ;;  4/5 dec e
	jp slot            ;; 10/11 jp z, stuffbit
dummy:
	ret                ;; 10/11

burn_more:
	ex (sp), ix         ; /25  make P8E loop 560 T
	ex (sp), ix         ; /25

	ex (sp), ix         ; /25
	ex (sp), ix         ; /25

	ex (sp), ix         ; /25
	ex (sp), ix         ; /25

	jp .+3             ;; /11
	jp .+3             ;; /11
	jp .+3             ;; /11
	jp .+3             ;; /11
	jp .+3             ;; /11

	ld a, (_logic)     ;; /14
	ld a, # 0          ;; /8

	ret                ;; /11

