
; hopeless with no input capture.

; RXAF -> LPF -> comparator -> PB.4
;
; TMR_MT is counting down 0 ... FFFF, 0 ...
;
;  Timer 2   1 / 4032000 = 248 nsec
;  ax25      1 / 1200    = 833 usec
;  2500 Hz half-cycle    = 200 usec
;  2200 Hz half-cycle    = 227 usec
;
;  1 bits count  3360 0x0D20  0 only
;  2 bits count  6720 0x1A40  10
;  3 bits count 10080 0x2760  110
;  4 bits count 13440 0x3480  1110
;  5 bits count 16800 0x41A0  11110
;  6 bits count 20160 0x4EC0  111110     stuffing.
;  7 bits count 23520 0x5BE0  1111110    that is a flag.
;  8 bits count 26880 0x6900  1111111... and this an abort.
;
;  suffices to check just the msbyte of the 16 bit count

	.module afsk_rx

	.include "iomap.inc"

PB_AFSK = 0x10               ; PB.4 was no connect

_1700   = TMR_CLK / 1700 / 2     ; half-cycle between 1200 and 2200 Hz

_0      = 3 * TMR_CLK / 1200 / 2 / 256 ; 1.5 bittimes max
_10     = 5 * TMR_CLK / 1200 / 2 / 256 ; 2.5
_110    = 7 * TMR_CLK / 1200 / 2 / 256 ; 3.5
_1110   = 9 * TMR_CLK / 1200 / 2 / 256 ; 4.5
_11110 = 11 * TMR_CLK / 1200 / 2 / 256 ; 5.5
_11111 = 13 * TMR_CLK / 1200 / 2 / 256 ; 6.5
_FLAG  = 15 * TMR_CLK / 1200 / 2 / 256 ; 7.5

	.area _DATA

pio_b_icw: .byte 0x97         ; IEN OR LOW MASK
bufptr:    .word _buffers

	.area _BSS

t_old:  .ds 2    ; previous zero cross
t_zero: .ds 2    ; when tone changed
tone:   .ds 1    ; in the msbit

	.area _CODE

_afsk_rx_off::

	rst DISABLE         ; funny, but according to folklore, necessary

	ld a, # 0x03        ; INTERRUPT DISABLE
	out (PIO_B_CTRL), a

	jp _restore         ; RET


_afsk_rx_on::
	ld hl, (bufptr)     ; current buffer rewound
	ld l, # 0
	ld (hl), # 0x7F     ; after 8 RRs, zero goes to CY
	ld (bufptr), hl

	ld a, # TMR_SEL_MT | TMR_BOTH | TMR_INTTC
	out (TMR_CTRL), a
	xor a
	out (TMR_MT), a     ; count down from 0x10000
	out (TMR_MT), a     ; after 16 msec, OUT rises and stays set

	ld a, # 0x83        ; INTERRUPT ENABLE
	out (PIO_B_CTRL), a
	ret


pio_b_int::              ; 19T (minimum)
	ex af, af'           ;  4T
	exx                  ;  4T

	ld hl, # pio_b_icw   ; 10T for next zero-cross,
	ld a, (hl)           ;  7T
	xor a, # 0x20        ;  7T toggle active level between LOW/HIGH
	ld (hl), a           ;  7T
	out (PIO_B_CTRL), a  ; 11T
	ld a, # ~PB_AFSK     ;  7T
	out (PIO_B_CTRL), a  ; 11T what if ICW did not contain MASK ?

	ld a, # TMR_SEL_MT | TMR_LATCH     ;  7T
	out (TMR_CTRL), a    ; 11T
	in a, (TMR_MT)       ; 11T
	ld l, a              ;  4T
	in a, (TMR_MT)       ; 11T
	ld h, a              ;  4T
	ld bc, (t_old)       ; 20T how long half-cycle ?
	ld (t_old), hl       ; 16T
	sbc hl, bc           ; 15T t - t_old, a negative value (8254 counts down)

	ld a, l              ;  4T below or above 1700 Hz ?
	add a, # _1700 % 256 ;  7T
	ld a, h              ;  4T
	adc a, # _1700 / 256 ;  7T A.7 set accordingly

	ld hl, # tone        ; 10T
	xor (hl)             ;  7T tone changed ?
	jp p, out            ; 10T no.
	xor (hl)             ;  7T
	ld (hl), a           ;  7T

	ld hl, (t_zero)      ; 16T how long the previous tone was ?
	ld (t_zero), bc      ; 20T still t_old
	sbc hl, bc           ; 15T t_zero - t_old, pulse length (now positive value)

	ld a, h              ;  4T how many 1200 baud bittimes ?
	ld hl, (bufptr)      ; 16T

	cp # _0              ;  7T      A.7 is zero. last inserted bit
	jr c, 1$             ;  7T/12T  CY always "1" when jumping
	cp # _10             ;  7T
	jr c, 2$             ;  7T/12T
	cp # _110            ;  7T
	jr c, 3$             ;  7T/12T
	cp # _1110           ;  7T
	jr c, 4$             ;  7T/12T
	cp # _11110          ;  7T
	jr c, 5$             ;  7T/12T
	cp # _11111          ;  7T
	jr nc, flag_or_abort ;  7T/12T
	cpl                  ;  4T      A.7 is one. last inserted bit
5$:
	rr (hl)              ; 15T CY set while no "refill" needed
	jp c, 4$             ; 10T
	inc l                ;  4T overflows catched by crc only
	ld (hl), # 0x7F      ; 10T empty slot, reset marker
	ld (bufptr), hl      ; 16T
	scf                  ;  4T for next rotate
4$:
	rr (hl)              ; 15T "1" inserted by all but last RR (HL)
	jp c, 3$             ; 10T
	inc l                ; refill executed at most once per interrupt
	ld (hl), # 0x7F
	ld (bufptr), hl
	scf
3$:
	rr (hl)              ; 15T
	jp c, 2$             ; 10T
	inc l
	ld (hl), # 0x7F
	ld (bufptr), hl
	scf
2$:
	rr (hl)              ; 15T
	jp c, 1$             ; 10T
	inc l
	ld (hl), # 0x7F
	ld (bufptr), hl
1$:
	rla                  ;  4T last bit from A.7
	rr (hl)              ; 15T
	jp c, out            ; 10T
	inc l
	ld (hl), # 0x7F
	ld (bufptr), hl
out:
	ex af, af'           ;  4T
	exx                  ;  4T
	ei                   ;  4T
	reti                 ; 14T   ========= totalling at 620T now

flag_or_abort:
	cp # _FLAG             ; longer than flag, abort ?
	jr nc, rewind 
	ld a, (hl)             ; bit residue ?
	cp # 0x3F              ; 7F, after the leading "0" of a flag
	jr nz, rewind 
	ld a, l                ; aligned buffers, packet length is LO(pointer)
	cp # 7 + 7 + 1 + 2     ; at least "dst src control crc" needed
	jr c, rewind 

	ld l, # 0xFF
	ld (hl), a             ; XXX length

;	ld l, # 0                ; check destination[0]
;	ld a, (hl)
;	sub # '0 << 1            ; has to be 0-9 A-Z
;	cp  # ('Z - '0) << 1     ; quick check does not catch the 7 bytes between 9 and A
;	jr c, rewind

	inc h                  ; next buffer
	res 3, h               ; XXX cheap trick, D000 ... D800
rewind:
	ld l, # 0              ;  7T same buffer, rewound
	ld (hl), # 0x7F
	ld (bufptr), hl        ; 16T
	ex af, af'           ;  4T JP out
	exx                  ;  4T
	ei
	reti

