; XXX
	.module crash

	.include "iomap.inc"

	.IF P8x
CSR  = SIO_A_CSR
DATA = SIO_A_DATA
	.ENDIF

	.IF L8x
CSR  = SIO_B_CSR
DATA = SIO_B_DATA
	.ENDIF

	.area _CODE

out_XX:
	ld d, a             ; D keeps lsnybble
	rra
	rra
	rra
	rra
	call 1$
	ld a, d
1$:
	and # 0x0F
	cp # 10             ; make ascii
	sbc # 0x69
	daa
out_c:
	ld e, a             ; E keeps character
1$:
	in a, (CSR)
	and # RR0_TBE
	jr z, 1$            ; power off by hw in 1 second, if this sticks

	ld a, e
	out (DATA), a
	ret

out_s_1:
	call out_c
	inc bc
out_s:
	ld a, (bc)
	or a
	jr nz, out_s_1
	ret

sio_init:
	.byte 4, WR4_1STOPBIT | WR4_X16_CLK
	.byte 3, WR3_8BIT_RX | WR3_RX_ENB
	.byte 5, WR5_8BIT_TX | WR5_TX_ENB
	.byte 1, WR1_STATUS_VECTOR | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC

	sio_init_len = . - sio_init

_crash::
	di
	xor a
	out (DA_TPC), a

	; XXX sio (and timer_brg) must be reconfigured

	ld bc, # sio_init_len * 0x100 + CSR
	ld hl, # sio_init
	in a, (c)            ; WR0, please
	otir

	.IF L8x

	CLK = 16 * 9600
	BRG = (TMR_CLK + CLK/2) / CLK

	ld a, # TMR_SEL_BRG | TMR_BOTH | TMR_SQWAVE
	out (TMR_CTRL), a
	ld a, #< BRG
	out (TMR_BRG), a
	ld a, #> BRG
	out (TMR_BRG), a

	.ENDIF

	; SP to HL

	ld hl, # 0
	add hl, sp           ; old sp in HL

	ld sp, # 0           ; two words needed here

	; buzzer

	ld a, # TMR_SEL_MT | TMR_BOTH | TMR_SQWAVE
	out (TMR_CTRL), a

	ld a, # TMR_CLK / 333
	out (TMR_MT), a
	ld a, # TMR_CLK / 333 / 256
	out (TMR_MT), a

	; rx mute, sidetone, earphone only

	ld a, # O0_MTC | O0_INH | O0_MICM   ; no O0_AUDIOC or O0_CCIRC
	out (OUT_0), a

	; jp _boot    ; XXX for now

	ld bc, # ugh
	call out_s          ; ... sad to report

	ld a, h
	call out_XX         ; SP
	ld a, l
	call out_XX

	jr 3$               ; start with linefeed
1$:
	ld a, (hl)          ; byte from stack
	call out_XX
	inc hl

	ld a, l
	and # 15            ; 15 times "XX "
	ld a, # 0x20        ; blank
	jr nz, 2$           ; then "XX\n" once
3$:
	ld a, # 0x0A        ; linefeed
2$:
	call out_c
	ld a, h
	cp #>_ram           ; ram 0xC000 ... 0xFFFF
	jr nc, 1$

	ld a, # 0x0A
	call out_c
	call out_c          ; extra linefeeds to drain it
	call out_c

	jp _boot

ugh: .asciz "\nugh...\n"

