
; sine contains values 1...127,
; low + high tone is 2 ... 254
;
; pwm rate is 4032 kHz per 256 or 15750 Hz
; P8N loop must take 256 T,
; P8E loop must take 512 T
;
; timer mode is programmed to
; interrupt-on-terminal count, lsbyte only.
; tx (and sidetone) audio paths are open.
;
; P8E times include the 1 (or 2 if prefixed) waitstate

	.module dtmf

	.include "iomap.inc"
	.include "cfg.inc"

	.area _BSS

phacc_1: .ds 2
phacc_2: .ds 2
c_len:   .ds 2
ogain:   .ds 1

	.area _CODE

setup:
	ld a, (_cfg + dtmf_gain)
	ld hl, # ogain
	cp (hl)
	jr z, 2$
	ld (hl), a

	; full range in sine ok, pwm not higher than 256 timer clocks (is equal)

	ld e, a             ; gain
	ld d, # 0x40        ; center (sum of two tones)
	ld iy, # _dtmf_sine
	call make_sine
2$:
	ld hl, # 4        ; args: word ra1, word ra, byte c, word msec
	add hl, sp
	ld a, (hl)        ; A = c
	inc hl
	ld e, (hl)
	inc hl
	ld h, (hl)
	ld l, e           ; HL = msec (only used by dtmf_c)
;	ld d, h           ; DE = msec
;	srl d
;	rr e              ; 0.5
;	srl d
;	rr e              ; 0.25
	add hl, hl        ; 2
	add hl, hl        ; 4
	add hl, hl        ; 8
	add hl, hl        ; 16     close enough
;	sbc hl, de        ; 15.75
	ld (c_len), hl

	ld hl, # 0
	ld (phacc_1), hl
	ld (phacc_2), hl

	ld hl, # _dtmf_mt_tab  ; { inc1, inc2, inc1, inc2, ... }
	add a, a               ; table index, 0 ... 16 incl.
	add a, a               ; two words per index
	add a, l
	ld l, a
	jr nc, 1$
	inc h
1$:
	ld c, (hl)
	inc hl
	ld b, (hl)             ; BC phinc_1
	inc hl
	ld e, (hl)
	inc hl
	ld d, (hl)             ; DE phinc_2

	rst DISABLE            ; from now on.

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	ret

;----------------------------------------------------------------------

_dtmf_timer_mt_a::           ; (byte x);
	call setup
1$:
	ld hl, (phacc_2)         ; 16/17 high tone
	add hl, de               ; 11/12 acc += inc
	ld (phacc_2), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	ld a, (hl)               ;  7/8  sine[]

	; add a, a ; 4or5 for extreme "twist" XXX needs smaller "center"

	ld hl, (phacc_1)         ; 16/17 low tone
	add hl, bc               ; 11/12 acc += inc
	ld (phacc_1), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	add a, (hl)              ;  7/8  sine[]

	out (TMR_MT), a          ; 11/12 low + high tone
	out (WDOG), a            ; 11/12

	ld hl, (phacc_1)         ; 16/17 dummies
	ld hl, (phacc_1)         ; 16/17
	or a                     ;  4/5  to make P8N loop 256 T

	ld a, (_logic)           ; 13/14
	cp # 0                   ;  7/8
	call z, burn_more        ; 10/18 extra needed for P8E

	ld a, # WR0_RESET_ESCINT ;  7/8
	out (SIO_A_CSR), a       ; 11/12 unlatch, fresh values
	in a, (SIO_A_CSR)        ; 11/12
	and # SA_DA              ;  7/8  key still pressed ?
	jp z, 1$                 ; 10/11 pin inverted by SIO
done:
	; stop timer

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	call sio_a_esc_pickup    ; pick up changes XXX
	jp _restore              ; RET


;----------------------------------------------------------------------
; _identical_ to above,
; except SA_INT and jp nz  XXX can do better XXX

_dtmf_timer_mt_b::           ; (byte x)
	call setup
1$:
	ld hl, (phacc_2)         ; 16/17 high tone
	add hl, de               ; 11/12 acc += inc
	ld (phacc_2), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	ld a, (hl)               ;  7/8  sine[]

	; add a, a ; 4or5 for extreme "twist" XXX needs smaller "center"

	ld hl, (phacc_1)         ; 16/17 low tone
	add hl, bc               ; 11/12 acc += inc
	ld (phacc_1), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	add a, (hl)              ;  7/8  sine[]

	out (TMR_MT), a          ; 11/12 low + high tone
	out (WDOG), a            ; 11/12

	ld hl, (phacc_1)         ; 16/17 dummies
	ld hl, (phacc_1)         ; 16/17
	or a                     ;  4/5  to make P8N loop 256 T

	ld a, (_logic)           ; 13/14
	cp # 0                   ;  7/8
	call z, burn_more        ; 10/18 extra needed for P8E

	ld a, # WR0_RESET_ESCINT ;  7/8
	out (SIO_A_CSR), a       ; 11/12 unlatch, fresh values
	in a, (SIO_A_CSR)        ; 11/12
	and # SA_INT             ;  7/8  changes in keypad ?
	jp nz, 1$                ; 10/11 pin inverted by SIO

	jp done

	.IF 0

	avoid duplicate routines for CU53 and CU58AF

	ld hl, (phacc_1)         ; 16/17 dummies
	ld hl, (phacc_1)         ; 16/17
	or a                     ;  4/5  to make P8N loop 256 T
	and # SA_INT             ;  7/8  changes in keypad ?

	jr 1$                    ; /13  P8E only


	ld a, (ix)               ; 19/21 dummy to make P8N loop 256T

	ld hl, (_dtmf_mt_ex)     ; 16/17 exit condition in SIO_A_RR0
	and l                    ;  4/5  when these bits
	xor h                    ;  4/5  are this way

	add hl, hl               ; /12  P8E only

	.ENDIF

;----------------------------------------------------------------------
; Gosh, THIRD one!

_dtmf_timer_mt_c::           ; (byte x, word msec);
	call setup
1$:
	ld hl, (phacc_2)         ; 16/17 high tone
	add hl, de               ; 11/12 acc += inc
	ld (phacc_2), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	ld a, (hl)               ;  7/8  sine[]

	ld hl, (phacc_1)         ; 16/17 low tone
	add hl, bc               ; 11/12 acc += inc
	ld (phacc_1), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	add a, (hl)              ;  7/8  sine[]

	out (TMR_MT), a          ; 11/12 low + high tone
	out (WDOG), a            ; 11/12

	ld a, (_logic)           ; 13/14 dummies
	ld a, (_logic)           ; 13/14 to make P8N loop 256 T

	ld a, (_logic)           ; 13/14
	cp # 0                   ;  7/8
	call z, burn_more        ; 10/18 extra needed for P8E

	ld hl, (c_len)           ; 16/17
	ld a, h                  ;  4/5  0 is shortest, FFFF longest (4 sec)
	or l                     ;  4/5
	dec hl                   ;  6/7
	ld (c_len), hl           ; 16/17
	jp nz, 1$                ; 10/11

	; stop timer

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	jp _restore              ; RET

;----------------------------------------------------------------------

burn_more:                   ;     to make P8E loop 512 T
	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	jr 1$                    ; /13
1$:
	ret                      ; /11

;----------------------------------------------------------------------

	.area _BSS

ptr: .ds 2
spd: .ds 1

	.area _CODE

_dtmf_burst::             ; void dtmf_burst(char *s);
	pop de                ; args:
	pop hl                ; string s
	push hl
	push de

	xor a
	ld (_af_tx_timer), a

	ld (ptr), hl

	ld hl, (_cfg + dtmf_speed)  ; msec of pause
	ld a, l
	srl h                       ; to 0.8 seconds
	rra                         ; /= 4 is close enough
	srl h
	rra 
	inc h
	dec h
	jr z, 1$
	ld a, # 255           ; sorry, one second is max
1$:
	ld (spd), a

	call _timer_mt_off
	call _CCIRC_on

	ld hl, # sender
	jp af_tx_start


sender:
	ld hl, (ptr)
	ld a, (hl)
	or a
	jp z, af_tx_sent
	inc hl
	ld (ptr), hl

	cp # 0x21             ; blank ?
	jr c, 2$              ; skip tone

	sub #'0               ; ascii 0-9A-D*# (EF also ok) to binary
	cp # 10
	jr c, 1$              ; 0-9
	sub # 7
	cp # 16
	jr c, 1$              ; A-D (or A-F)
	cp #'* - ('0 + 7)
	ld a, # 0xE           ; *
	jr z, 1$
	ld a, # 0xF           ; # else.
1$:
	ld hl, (_cfg + dtmf_speed)  ; msec of tone
	push hl
	push af               ; byte c
	inc sp
	call _dtmf_timer_mt_c ; freezes time
	inc sp
	pop af
2$:
	ld hl, (ptr)
	ld a, (hl)            ; more ?
	or a
	jp z, af_tx_sent

	ld a, (spd)
	ret

