
typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

#define NULL      ((void *) 0)
#define __SIZE(r) (sizeof(r)/sizeof(*(r)))

#define isprint(c) ((byte)((c)-' ') < 96)
#define isdigit(c) ((byte)((c)-'0') < 10)
#define isupper(c) ((byte)((c)-'A') < 'Z'-'A'+1)

#define offsetof(t,m) ((word) &((t *)0)->m)

extern word
	TASK_0[],
	TASK_1[],
	TASK_2[];

extern void
	boot(),
	crash(),
	power_loss(),
	power_off(),

	disable(),
	restore(),
	enable(),
	hog(),
	yield(),
	sleep(),
	wake(void *),

	load_bbram(),
	save_bbram(),

	ctcss(word dHz),
	dcs(int code),
	dcs_go(),
	encoder_turnoff(),
	encoder_end(),

	pll_strobe(byte mask),
	pll_bits(byte mask, byte bits),
	cu_bit(byte nz),

	snap_squelch(),

	dtmf(byte),

	dtmf_burst(byte *),
	morse_send(byte *),
	ccir_send(byte *),
	af_tx_stop(),

	lights_on(),
	tx_led_on(),
	tx_led_off(),
	call_led_on(),
	call_led_off(),
	update_leds(),
	display_soon(),
	display(),
	nop();

extern dword
	dcs_frame(int code);

extern byte
    i2c_get(byte addr),
    i2c_put(byte addr, byte),
    i2c_write(byte addr, byte cc, byte *),
	read_ccir(),
	read_8870(),
	read_8888(),
	cointoss(),            /* randomish 0 or 1 */
	xlat(byte *, byte),
	*axaddr(byte *dst, byte *asc), /* returns dst + 7 */

	ptt_pressed(),
	off_hook();

extern const byte
	revnybble[],
	revbyte[],             /* fx databits go msbit first */
	version[];

word strlen  (byte *);
byte *strend (byte *);
byte *index  (byte *, word c);
byte *rindex (byte *, word c);

int strcmp   (byte *, byte *);

byte *strcpyx  (byte *dst, byte *src);
byte *strncpyx (byte *dst, byte *src, word bufsiz);

void bcopy (void *src, void *dst, word len);
void bzero (void *, word len);
byte bcmp  (void *, void *, word len);

void memcpy (void *dst, void *src, word len);
void memset (void *dst, word c, word len);

word crc_ccitt(byte *, word len);

void delay_usec(word usec);         /* uses 'logic' for cyclecount */

void store_word(void *, word);
void store_dword(void *, dword);
word  load_word(void *);
dword load_dword(void *);

void make_sine(byte *, byte gain, byte center);

#ifdef P8x

word sio_a_getc();
void sio_a_putc(word);
void sio_a_puts(byte *);

#endif

word sio_b_getc();
void sio_b_putc(word);
void sio_b_puts(byte *);

extern byte hexdig[];

byte nybble(char);

byte  atob (byte *);
word  atow (byte *);
dword atodw(byte *);

dword atodw_seeded(byte *, dword);

byte
	ulmoddiv(dword *, byte),
	umoddiv(word *, byte);

word
	sqrt(dword),
	abs(word);

dword
	log1_08(dword),
	log1_002(dword),
	labs(dword),
	divide_dwdww(dword, word),
	divide_dwdwdw(dword, dword);

/* different views into 32.24 time, seconds */

extern dword timex();    /* seconds 32.0 */
extern word  timef();    /* seconds 8.8 */

extern word timef_mark(word interval);
extern byte timef_passed(word mark);

extern word time_mark(word interval);
extern byte time_passed(word mark);

/* enumeration order is important */

extern byte logic;
enum {
	LOGIC_P8E,   /* only 8 MHz */
	LOGIC_P8N,   /* cfg must be moved to non-volatile ram area */
	LOGIC_L8M,   /*                                eeprom area */
	LOGIC_L8TM,  /* & 3 == 3 */
};

extern byte
	ad_rssi_peak_timo,
	ad_rssi_peak,        /* held for one second */
	ad_rssi,
	ad_sql,
	ad_batt,
	ad_tpc,
	ad_fpm,
	ad_rpm,
	ad_tp4,
	ad_in7,

	pio_a,
	pio_a_ddr,
	pio_b,
	pio_b_ddr,

	receiving,
	transmitting,
	transmit_state,
	transmit_tail, afc,
	alert,
	setup_active,
	dpy_XXX;

extern word
	ad_batt_millivolt();

extern dword
	rx_freq,
	tx_freq;

extern signed char
	doppler;

void RFC_update();
byte RFC_get();
void RFC_set(byte rfc);

word centibel(byte rssi);

enum { FSK, CCIR, DTMF, MORSE }; /* buf type */

struct buf {
	struct buf *next;
	byte len;
	byte type;
	byte data[60]; /* some preamble, sync word, data, crc */
};

extern struct buf
	*fx_tx_head, **fx_tx_tail,
	*deq_incoming(),
	*buf_free,
	*buf_alloc();

extern void
	enq_incoming(struct buf *),
	buf_relse(struct buf *),
	nmea_input(byte);

extern int
	nmea_split(byte *);

/* cu53 or cu58af keypad or programmed time */

extern void
	dtmf_timer_mt_a(byte),
	dtmf_timer_mt_b(byte),
	dtmf_timer_mt_c(byte, word msec);

extern byte dtmf_sine[];

/* unsigned number from right to left,
 * no leading zeroes,
 * &msdigit returned */

extern byte
	*format_hex(byte *endp, word),
	*format_word(byte *, word),
	*format_short(byte *, short),
	*format_dword(byte *, dword),
	*format_long(byte *, dword);

