
/* for L8TM
 *
 * 0000 ... 7FFF rom0, 2 slices
 * 8000 ... BFFF rom1, 4 slices
 * C000 ... DFFF bbram, 2 slices
 * C000 ... FFFF work ram
 *
 * iomap_L8TM.inc is a subset of this file */

IO_ADDR(PIO_A_DATA, 0x00)
IO_ADDR(PIO_B_DATA, 0x01)
IO_ADDR(PIO_A_CTRL, 0x02)
IO_ADDR(PIO_B_CTRL, 0x03)
IO_ADDR(SIO_A_DATA, 0x10)
IO_ADDR(SIO_B_DATA, 0x11)
IO_ADDR(SIO_A_CSR,  0x12)
IO_ADDR(SIO_B_CSR,  0x13)
IO_ADDR(TMR_BRG,    0x20)
IO_ADDR(TMR_LPF,    0x21)
IO_ADDR(TMR_MT,     0x22)
IO_ADDR(TMR_CTRL,   0x23)
IO_ADDR(DA_RFC,     0x30) /* or vice versa ? */
IO_ADDR(DA_TPC,     0x40)
IO_ADDR(OUT_1,      0x50) /* == OUT_2 */
IO_ADDR(OUT_2,      0x50) /* OUT0 in manual, works as P8x OUT1 and OUT2 */
IO_ADDR(OUT_0,      0x60) /* OUT1 in manual, works as P8x OUT0 */
IO_ADDR(BANK,       0x70) /* 0xE0? 0x18? 0x10 blocks RxCA? */
IO_ADDR(AD_RSSI,    0x80)
IO_ADDR(AD_FPM,     0x81) /* aka TPL */
IO_ADDR(AD_BATT,    0x82)
IO_ADDR(AD_RFC,     0x83)
IO_ADDR(AD_RPM,     0x84) /* aka TRPL */
IO_ADDR(AD_TPC,     0x85)
IO_ADDR(AD_SQL,     0x86) /* SQL (unpopulated L8TM circuit) ? */
IO_ADDR(AD_IN7,     0x87) /* R8C SQL ? */
IO_ADDR(WDOG,       0x90)

/*
 * IN0 = RSSI
 * IN1 = TPL
 * IN2 on vastusjaolla V+:aan
 * IN3 = RFC
 * IN4 = +TRPL
 * IN5 = TPC
 * IN6 = SQL
 * IN7 on johdotettu RP-liittimen 7-nastaan,
 *     eli R8C +kohinasalpatietoon, ei kytss niss
 */

extern byte
	out_0,
	out_1, /* out_1 == out_2 */
	out_2,
	bank;

#define O0_VOLUME 0x07 /* speaker volume */
#define O0_INH    0x08 /* speaker off */
#define O0_AUDIOC 0x10
#define O0_CCIRC  0x20
#define O0_MTC    0x40
#define O0_MICM   0x80

/* mimic P8x */

#define O2_DP     0x01 /* aka O1_SD  */
#define O2_CLK    0x02 /* aka O1_CLK */
#define O1_SRE    0x04
#define O1_STE    0x08
#define O2_CS1    0x10
#define O2_CS2    0x20
#define O1_TPS    0x40 /* aka RFS, same pin */
#define O1_TXOFF  0x80
#define O1_TXON   0x00 /* dummy */

/* XXX if bastard ...
#define O1_TXOFF  0x00
#define O1_TXON   0x80 this way */

#define O1_CLK O2_CLK
#define O1_SD  O2_DP


#define BNK_ROM1A14   0x01
#define BNK_ROM1A15   0x02
#define BNK_ROM0A15   0x04
#define BNK_RXCA      0x18   /* sio rxca from */
#define BNK_RXCA_RXC  0x08   /*  fx419 rxc */
#define BNK_RXCA_TMR1 0x10   /*  timer 1 output */
/* two msbits are no connect */

#define SA_ONHOOK RR0_DCD
#define SA_DA     RR0_CTS	/* DA or I2C /INT */

#define SA_nRXE   WR5_DTR   /* fsk modem RXE. inversion in sio accounted for */
#define SA_TXE    WR5_RTS   /* fsk modem /TXE */

#define SB_PTT    RR0_CTS
#define SB_LOCAL  RR0_SYNC
#define SB_DMIDLE RR0_DCD	/* 1 means MBUS idle */

#define SB_MON    WR5_DTR   /* Master ON pulse, a/c coupled */


#define PA_CLK2    0x01  /* 1968.75 Hz square wave      */
#define PA_SCE     0x02  /* synth control reg strobe or vco buffer /enable */
#define PA_WDR     0x04  /* watchdog was activated      */
#define PA_PWR     0x08  /* on/off button, 1 means off  */
#define PA_CCIR    0xF0  /* FX003 data                  */

#define PB_RAMA12 0x01	/* which half of config    out               */
#define PB_EXIN1  0x02	/* aka /POR                out 0 if SERV     */
#define PB_EXIN2  0x04  /* aka /IGN aka /EMG       out 0 if GPio2b 0 */
#define PB_DCU    0x08	/* I2C SDA for CU58AF      out 0 if SDA 0    */
#define PB_AVAIL  0x10  /* no connect  ROM S/L ?   in                */
#define PB_SMEM   0x20  /* config ram slices if 0  out               */
#define PB_EXAL   0x40	/*                         out               */
#define PB_PWROFF 0x80	/*                         out               */

/* all gates are +5v */

#define TMR_SEL_BRG  0x00
#define TMR_SEL_LPF  0x40
#define TMR_SEL_MT   0x80

IO_ADDR(MT8888_WD,  0xC0) /* write data,    RS0=0, A1=0 */
IO_ADDR(MT8888_WC,  0xC1) /* write control, RS0=1, A1=0 */
IO_ADDR(MT8888_RD,  0xC2) /* read data,     RS0=0, A1=1 */
IO_ADDR(MT8888_RS,  0xC3) /* read status,   RS0=1, A1=1 */

