
/* for P8E/P8N
 *
 * 0000 ... 7FFF rom0 bottom
 * 8000 ... BFFF rom0 top, 2 slices or rom1, multiple slices
 * C000 ... DFFF bbram, 2 slices
 * C000 ... FFFF work ram
 *
 * iomap_P8x.inc is a subset of this file */

IO_ADDR(PIO_A_DATA, 0x00)
IO_ADDR(PIO_B_DATA, 0x01)
IO_ADDR(PIO_A_CTRL, 0x02)
IO_ADDR(PIO_B_CTRL, 0x03)
IO_ADDR(SIO_A_DATA, 0x10)
IO_ADDR(SIO_B_DATA, 0x11)
IO_ADDR(SIO_A_CSR,  0x12)
IO_ADDR(SIO_B_CSR,  0x13)
IO_ADDR(TMR_LPF,    0x20)
IO_ADDR(TMR_MT,     0x21)
IO_ADDR(TMR_2,      0x22) /* unused */
IO_ADDR(TMR_CTRL,   0x23)
IO_ADDR(DA_RFC,     0x30)
IO_ADDR(DA_TPC,     0x40)
IO_ADDR(AD_RSSI,    0x50)
IO_ADDR(AD_SQL,     0x51)
IO_ADDR(AD_BATT,    0x52)
IO_ADDR(AD_TPC,     0x53)
IO_ADDR(AD_FPM,     0x54)
IO_ADDR(AD_RPM,     0x55)
IO_ADDR(AD_TP4,     0x56)
IO_ADDR(AD_IN7,     0x57)
IO_ADDR(OUT_0,      0x60)
IO_ADDR(OUT_1,      0x70)
IO_ADDR(OUT_2,      0x80)
IO_ADDR(WDOG,       0x90)
IO_ADDR(FX429_DATA, 0xA2)
IO_ADDR(FX429_CSR,  0xA3)
IO_ADDR(CSMEM,      0xB0) /* last 3 only on P8E */
IO_ADDR(CSPROM,     0xC0)
IO_ADDR(CSKEY,      0xD0)

IO_ADDR(MT8888_WD,  0xC0) /* write data,    RS0=0, A1=0 */
IO_ADDR(MT8888_WC,  0xC1) /* write control, RS0=1, A1=0 */
IO_ADDR(MT8888_RD,  0xC2) /* read data,     RS0=0, A1=1 */
IO_ADDR(MT8888_RS,  0xC3) /* read status,   RS0=1, A1=1 */

extern byte
	out_0,
	out_1,
	out_2,
	fx429_cmd;

#define O0_VOLUME 0x07 /* speaker volume */
#define O0_INH    0x08 /* speaker off */
#define O0_AUDIOC 0x10
#define O0_CCIRC  0x20
#define O0_MTC    0x40
#define O0_MICM   0x80

#define O1_SRE    0x01
#define O1_SCE    0x02
#define O1_STE    0x04
#define O1_CLK    0x08
#define O1_SD     0x10
#define O1_RAS    0x20
#define O1_TPS    0x40
#define O1_TXOFF  0x80
#define O1_TXON   0x00 /* dummy */

#define O2_ROMA14 0x01 /* select which ROM bank is at 0x8000...0xC000 */
#define O2_ROMA15 0x02
#define O2_ROM0   0x04 /* 2 banks from ROM0, if '1' */
#define O2_SMEM   0x08 /* ROMA16 on P8E (CSMEM elsewhere) */
#define O2_CS1    0x10
#define O2_CS2    0x20
#define O2_CLK    0x40 /* I2C SCL for CU58AF */
#define O2_DP     0x80

#define SA_KKINT  RR0_DCD
#define SA_DA     RR0_CTS	/* DA or I2C /INT */

#define SB_PTT    RR0_CTS
#define SB_LOCAL  RR0_SYNC
#define SB_DMIDLE RR0_DCD	/* 1 means MBUS idle */

#define SB_MON    WR5_DTR   /* Master ON pulse, a/c coupled */


#define PA_CLK2    0x01  /* 1968.75 Hz square wave      */
#define PA_OFFHOOK 0x02  /* state of HOOK pin, 0 onhook */
#define PA_WDR     0x04  /* watchdog was activated      */
#define PA_PWR     0x08  /* on/off button, 1 means off  */
#define PA_CCIR    0xF0  /* FX003 data                  */

#define PB_RAMA12 0x01	/* S#/L ? P8E: no effect   out               */
#define PB_EXIN1  0x02	/* aka /POR                out 0 if SERV     */
#define PB_EXIN2  0x04  /* aka /IGN aka /EMG       out 0 if GPio2b 0 */
#define PB_DCU    0x08	/* I2C SDA for CU58AF      out 0 if SDA 0    */
#define PB_RXON   0x10  /* /RXON ? wassit ?        out               */
#define PB_TMR0   0x20  /* from timer 2 output     in (old CTCSS in) */
#define PB_EXAL   0x40	/*                         out               */
#define PB_PWROFF 0x80	/*                         out               */

#define FX429_TXENB  0x01
#define FX429_TXPAR  0x02
#define FX429_RXENB  0x04
#define FX429_RXFMT  0x08
#define FX429_TIMER  0xF0

#define FX429_RXRDY  0x01
#define FX429_RXTRUE 0x02
#define FX429_DCD    0x04
#define FX429_TXRDY  0x08
#define FX429_TXIDL  0x10
#define FX429_TMRINT 0x20
#define FX429_SYNC   0x40
#define FX429_SYNT   0x80

/* all gates are +5v */

#define TMR_SEL_LPF 0x00
#define TMR_SEL_MT  0x40
#define TMR_SEL_2   0x80

