
; AF BC DE HL IY can be freely used
; IX is frame pointer
;
; args in stack, byte (yes!), word or dword size
; return value in L, HL or DEHL
;
; zero count means 0x10000 in many cases

	.module util

	.include "iomap.inc"
	.include "cfg.inc"

	.area _CODE

;----------------------------------------------------------------------

	.IF 0          ; builtin with SDCC

_iow::             ; void iow(port, byte);
	pop hl
	pop bc
	push bc

	out (c), b
	jp (hl)

_ior::             ; byte ior(port);
	pop iy
	pop bc
	push bc

	in l, (c)
	jp (iy)

	.ENDIF

;----------------------------------------------------------------------
; element headers { next, prev, ... } must not cross pages

	.IF 0

unlink::            ; remove element HL
	ld c, (hl)
	inc l
	ld b, (hl)      ; BC next
	inc l

	ld a, (hl)
	inc l
	ld h, (hl)
	ld l, a         ; HL and HA prev

	ld (hl), c
	inc l
	ld (hl), b      ; prev->next = next

	inc c
	inc c
	ld (bc), a
	inc c
	ld a, h
	ld (bc), a      ; next->prev = prev
	ret

link::              ; after element HL (pos), append element DE (new)
	ld c, (hl)
	ld (hl), e
	inc l
	ld b, (hl)      ; BC pos->next
	ld (hl), d      ; pos->next = new
	dec l

	ex de, hl       ; HL new, DE pos
	ld a, l         ; HA new

	ld (hl), c
	inc l
	ld (hl), b      ; new->next = pos->next
	inc l
	ld (hl), e
	inc l
	ld (hl), d      ; new->prev = pos

	inc c
	inc c
	ld (bc), a
	inc c
	ld a, h
	ld (bc), a      ; new->next->prev = new
	ret

_deq::
	pop af
	pop hl           ; { head, tail }
	push hl
	push af
deq::
	di

	ld e, (hl)
	inc hl
	ld d, (hl)       ; DE first

	ld a, d
	or e
	jr z, 1$         ; q was empty

	dec hl           ; HL head
	ld c, l
	ld b, h          ; BC head
	ld a, (de)
	ld (hl), a
	inc hl
	inc de
	ld a, (de)       ; head = first->next
	ld (hl), a
	dec de           ; DE first
	or a             ; XXX checking msbyte only (00?? is rom)
	jr nz, 1$        ; did not become empty
	inc hl           ; HL tail
	ld (hl), c
	inc hl
	ld (hl), b       ; tail = &head
1$:
	ex de, hl        ; HL first

	jp di_undo

_enq::
	pop af
	pop hl           ; { head, tail }
	pop bc           ; { next, ... }
	push bc
	push hl
	push af
enq::
	xor a
	ld (bc), a
	inc bc
	ld (bc), a       ; new->next = NULL
	dec bc

	inc hl
	inc hl           ; HL &tail

	di

	ld e, (hl)
	ld (hl), c
	inc hl
	ld d, (hl)       ; DE last
	ld (hl), b       ; tail = new

	ex de, hl        ; HL last (or &head)
	ld (hl), c
	inc hl
	ld (hl), b       ; last->next = new

	jp di_undo

	.ENDIF

;----------------------------------------------------------------------

	.area _DATA

_incoming_head: .dw 0
_incoming_tail: .dw _incoming_head

	.area _CODE

_enq_incoming::
	pop af
	pop de
	push de
	push af

	xor a
	ld (de), a
	inc de
	ld (de), a               ; p->next = NULL
	dec de

	di

	ld hl, (_incoming_tail)  ; tail = *tail = p
	ld (_incoming_tail), de
	ld (hl), e
	inc hl
	ld (hl), d

	jp di_undo

_deq_incoming::
deq_incoming::

	di

	ld hl, (_incoming_head)  ; p ?
	ld a, h
	or l
	jp z, di_undo            ; nothing. NULL, Z

	xor a
	ld e, (hl)
	ld (hl), a
	inc hl
	ld d, (hl)               ; next
	ld (hl), a               ; p->next = NULL
	dec hl
	ld (_incoming_head), de  ; head updated.
	ld a, d
	or e
	jp nz, di_undo           ; still something. ok, NZ

	ld de, # _incoming_head
	ld (_incoming_tail), de  ; empty: tail = &head

	inc a                    ; return NZ
	jp di_undo

;----------------------------------------------------------------------

_abs::             ; word abs(int);
	pop de
	pop hl
	push hl
	push de

abs:
	ld a, h
	or a
	ret p          ; already positive
	cpl
	ld h, a
	ld a, l
	cpl
	ld l, a
	inc hl
	ret

_labs::            ; dword labs(long);
	pop bc
	pop hl
	pop de         ; DEHL
	push de
	push hl
	push bc

	bit 7, d
	ret z          ; already positive
lneg::
	xor a
	ld c, a        ; constant 0
	sub l
	ld l, a
	ld a, c
	sbc h
	ld h, a
	ld a, c
	sbc e
	ld e, a
	ld a, c
	sbc d
	ld d, a
	ret

neg::
	xor a
	sub l
	ld l, a
	ld a, # 0
	sbc h
	ld h, a
	ret

;----------------------------------------------------------------------

_log1_002::           ; dword log1_002(dword);
	pop af
	pop hl
	pop de
	push de
	push hl
	push af

	call log2p1

	ld bc, # 44406    ; log2(1.002) << 7
	call mul_dehl_bc
	ld a, e           ; drop 16 + 7 == 24 - 1 lsbits
	ld l, d
	ld h, c
	ld e, b
	ld d, # 0
	rla
	adc hl, hl
	rl e
	rl d
	ret

_log1_08::            ; dword log1_08(dword);
	pop af
	pop hl
	push hl
	push af

	call log2p1

	ld bc, # 36890    ; log2(1.08) << 12
	call mul_dehl_bc
	ld a, d           ; drop 16 + 12 == 32 - 4 lsbits
	ld l, c
	ld h, b
	ld de, # 0

	ld b, # 4
1$:
	rla
	adc hl, hl
	rl e
	djnz 1$
	ret

;----------------------------------------------------------------------

log2p1:
	inc l
	jr nz, log2
	inc h
	jr nz, log2
	inc e
	jr nz, log2
	inc d
	jr nz, log2

	ld e, # 31
	dec hl            ; .9999
	ret

log2:
	ld a, # 32        ; integer part
1$:
	inc d
	dec d
	jr nz, 2$         ; DEH, msbit in D
	ld d, e
	ld e, h
	ld h, l
	ld l, # 0         ; DEHL <<= 8
	sub # 8
	jr nz, 1$
	ret               ; 0.0
2$:
	ld c, a

	ld l, h
	ld h, e
	ld a, d           ; AHL
3$:
	dec c             ; integer part
	add hl, hl
	rla
	jr nc, 3$         ; align to AH, find and drop msbit

	ld e, h           ; lsbits
	ld l, a
	ld h, #> _log_tab
	ld b, (hl)        ; tab[i]
	inc l             ; if wrap to 0, Z set and A will be 0 from [0]
	ld a, (hl)        ; tab[i+1]
	jr z, 4$          ; AE = E
	sub b             ; tab[i+1] - tab[i]
	ld h, a
	call mul_h_e      ; A BC E kept
	ld e, l
	ld a, h
4$:
	add b             ; tab[i]
	ld h, a
	ld l, e

	ld e, c           ; integer part
	ld d, # 0         ; E.HL
	ret

;----------------------------------------------------------------------

_sqrt::              ; word sqrt(dword n);
	pop af
	pop hl
	pop de
	push de
	push hl
	push af
sqrt::
	ld bc, # 0xFFFF  ; guess x, be sure not to guess too low
	ld a, d
	or a
	jr nz, 1$
	ld b, # 0x0F     ; 0FFF
	ld a, e
	or a
	jr nz, 1$
	ld b, a          ; 00FF
	ld a, h
	or a
	jr nz, 1$
	ld c, # 0x0F     ; 000F
	ld a, l
	or a
	ret z            ; 0
1$:
	rla              ; input msbyte >= 0x40 ?
	jr c, 2$
	rla
	jr c, 2$
	rr b             ; guess /= 2
	rr c
	jr 1$            ; until nonzero pair of bits found.
3$:
	pop hl           ; refresh DEHL
	pop de
2$:
	push de          ; keep n on stack
	push hl

	push bc          ; Newton; x = (n / x + x) / 2
	call div_dehl_bc
	pop de
	add hl, de
	rr h
	rr l
	ld c, l          ; new x
	ld b, h
	or a
	sbc hl, de       ; while new x < old x
	jr c, 3$

	pop af           ; purge n
	pop af
	ex de, hl        ; old x is better (or same), forget last iteration
	ret

;----------------------------------------------------------------------

_strcmp::          ; short strcmp(byte *, byte *);
	pop bc
	pop de
	pop hl
	push hl
	push de
	push bc
1$:
	ld a, (de)
	cp (hl)
	jr nz, 2$      ; difference
	inc de
	inc hl         ; CPI would be slower
	or a
	jp nz, 1$      ; while same, but not \0

	ld l, a        ; same. return 0
	ld h, a
	ret
2$:
	sub (hl)
	ld l, a        ; LSByte of result
	sbc a, a       ; 0 or FF
	ld h, a
	ret

;----------------------------------------------------------------------

_index::           ; byte *index(byte *, word c);
	pop de
	pop hl
	pop bc         ; byte in c
	push bc
	push hl
	push de
1$:
	ld a, (hl)
	cp c
	ret z          ; HL points to the byte, can be \0
	inc hl
	or a
	jp nz, 1$

	ld l, a        ; NULL
	ld h, a
	ret

;----------------------------------------------------------------------

_rindex::          ; byte *rindex(byte *, word c);
	pop hl
	pop de
	pop bc         ; byte in c, \0 (silly) ok
	push bc
	push de
	push hl

	ld hl, # 0     ; NULL
1$:
	ld a, (de)
	cp c
	jp nz, 2$
	ld l, e
	ld h, d        ; could be this
2$:
	inc de
	or a
	jp nz, 1$

	ret            ; NULL or not, in HL

;----------------------------------------------------------------------

_memchr::          ; byte *memchr(void *, word c, word len);
	pop af
	pop hl
	pop de         ; byte in e
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, e
	cpir
	dec hl         ; back to the byte
	ret z          ; if found

	ld hl, # 0     ; else NULL
	ret

;----------------------------------------------------------------------

_strend::          ; byte *strend(byte *);
	pop de
	pop hl         ; string
	push hl
	push de

	xor a
	ld c, a        ; BC 0
	ld b, a
	cpir
	dec hl         ; back over \0
	ret            ; end in HL

;----------------------------------------------------------------------

_strlen::          ; word strlen(byte *);
	pop de
	pop hl         ; string
	push hl
	push de

	xor a          ; clear carry
	ld c, a        ; and BC
	ld b, a
	cpir
	ld hl, # -1
	sbc hl, bc     ; carry still clear
	ret            ; length in HL

;----------------------------------------------------------------------

_strcpyx::         ; byte *strcpyx(byte *dst, byte *src);
	pop bc
	pop de
	pop hl         ; src
	push hl
	push de
	push bc

	xor a
1$:
	cp (hl)
	ldi            ; keeps Z
	jp nz, 1$

	dec de         ; back over \0
	ex de, hl
	ret            ; end of dst in HL

;----------------------------------------------------------------------

_strncpyx::        ; byte *strncpyx(byte *dst, byte *src, word sz);
	pop af
	pop de
	pop hl
	pop bc
	push bc
	push hl
	push de
	push af

	dec bc         ; leave room for \0
	xor a
1$:
	cp (hl)        ; \0 ?
	jr z, 2$
	ldi
	jp pe, 1$      ; while sz
2$:
	ld (de), a
	ex de, hl
	ret            ; end of dst in HL

;----------------------------------------------------------------------

_memset::          ; void memset(void *, word c, word len);
	pop af
	pop hl
	pop de         ; word c
	pop bc         ; len
	push bc
	push de
	push hl
	push af

	ld (hl), e
	dec bc
	ld a, c
	or b
	ret z          ; len 1

	ld e, l
	ld d, h
	inc de
	ldir
	ret

;----------------------------------------------------------------------

_bzero::           ; void bzero(void *, word len);
	pop de
	pop hl
	pop bc         ; len
	push bc
	push hl
	push de        ; ra

	ld a, c
	or b
	ret z          ; len 0

	ld (hl), # 0
	dec bc
	ld a, c
	or b
	ret z          ; len 1

	ld e, l
	ld d, h
	inc de
	ldir
	ret

;----------------------------------------------------------------------

_memcpy::          ; void memcpy(void *dst, void *src, word len);
	pop af
	pop de
	pop hl
	pop bc
	push bc
	push hl
	push de
	push af
	ldir
	ret

_bcopy::           ; void bcopy(void *src, void *dst, word len);
	pop af
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, c
	or b           ; clears carry
	ret z          ; len 0

	sbc hl, de     ; carry if src < dst. not bothering with RET Z
	add hl, de
	jp c, 1$       ; copy backwards
	ldir
	ret
1$:
	add hl, bc
	ex de, hl
	add hl, bc
	ex de, hl
	dec hl
	dec de
	lddr
	ret

;----------------------------------------------------------------------

_bcmp::            ; byte/bool bcmp(void *, void *, word len);
	pop af
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, c
	or b
	jr z, 2$       ; len 0, same
1$:
	ld a, (de)
	inc de
	cpi
	jr nz, 3$      ; different
	jp pe, 1$      ; while len
2$:
	ld l, # 0
	ret            ; same.
3$:
	ld l, # 1
	ret            ; different

;----------------------------------------------------------------------

add_hl_a::
	add a, l
	ld l, a
	ret nc
	inc h
	ret

;----------------------------------------------------------------------

_xlat::            ; byte xlat(dbyte *tab, byte in);
	pop af
	pop hl
	pop bc
	push bc
	push hl
	push af

	dec hl         ; compensate for preincrement
1$:
	inc hl
	ld a, (hl)     ; if this
	inc hl
	cp c
	jr z, 2$
	or a           ; or not listed
	jr nz, 1$
2$:
	ld l, (hl)     ; replace with this
	ret

;----------------------------------------------------------------------

_atodw_seeded::
	pop af
	pop iy        ; numstring
	pop hl        ; dword
	pop de        ;  seed
	push de
	push hl
	push iy
	push af

	jr atodw_continue

_atodw::
	pop de
	pop iy
	push iy
	push de

atodw:
	ld de, # 0       ; DEHL = 0
	ld hl, # 0

atodw_continue:
	ld a, (iy)       ; while isdigit
	sub # '0
	cp # 10
	ret nc           ; nondigit, out, no carry

	inc iy

	ld c, l          ; DEHL *= 10
	ld b, h
	push de          ; msword to stack
	add hl, hl
	rl e
	rl d             ; *= 2
	jr c, overflow_  ; remember to purge it
	add hl, hl
	rl e
	rl d             ; *= 4
	jr c, overflow_
	add hl, bc
	pop bc           ; msword from stack
	ex de, hl
	adc hl, bc
	ex de, hl        ; *= 5
	jr c, overflow
	add hl, hl
	rl e
	rl d             ; *= 10
	jr c, overflow

	add a, l         ; DEHL += A
	ld l, a
	jr nc, atodw_continue
	inc h
	jr nz, atodw_continue
	inc e
	jr nz, atodw_continue
	inc d
	jr nz, atodw_continue
overflow:
	ld de, # 0xFFFF
	ld hl, # 0xFFFF
	ret

overflow_:
	pop hl
	jr overflow

_atow::
	pop de
	pop iy
	push iy
	push de

	call atodw       ; string at iy
	ld a, e
	or d
	ret z
	jr overflow

_atob::
	pop de
	pop iy
	push iy
	push de

	call atodw       ; string at iy
	ld a, h
	or e
	or d
	ret z
	jr overflow

;----------------------------------------------------------------------
; HL = H * E
; keeps A BC E
; zeroes D

mul_h_e:
	ld l, # 0
	ld d, l

	sla h
	jr nc, 1$
	add hl, de
1$:
	add hl, hl
	jr nc, 2$
	add hl, de
2$:
	add hl, hl
	jr nc, 3$
	add hl, de
3$:
	add hl, hl
	jr nc, 4$
	add hl, de
4$:
	add hl, hl
	jr nc, 5$
	add hl, de
5$:
	add hl, hl
	jr nc, 6$
	add hl, de
6$:
	add hl, hl
	jr nc, 7$
	add hl, de
7$:
	add hl, hl
	ret nc
	add hl, de
	ret

;----------------------------------------------------------------------
; AHL = A * HL
; keeps B
; 50 ... 75 usec

mul_hl_a::
	ex de, hl       ;  4     A and DE

mul_de_a:
	ld l, # 0       ;  7
	ld h, l         ;  4     HL 0
	or a            ;  4
	ret z           ; 11 5   quick result, AHL 0

	ld c, l         ;  4     C 0 for faster ADC A, # 0

	add a, a        ;  4     1st iteration pared down
	jr nc, 1$       ;  7 12
	add hl, de      ; 11
1$:
	add hl, hl      ; 11     AHL <<= 1
	rla             ;  4
	jr nc, 2$       ;  7 12
	add hl, de      ; 11     AHL += CDE
	adc c           ;  4
2$:
	add hl, hl      ; one iteration 27 or 37 T
	rla
	jr nc, 3$
	add hl, de
	adc c
3$:
	add hl, hl
	rla
	jr nc, 4$
	add hl, de
	adc c
4$:
	add hl, hl
	rla
	jr nc, 5$
	add hl, de
	adc c
5$:
	add hl, hl
	rla
	jr nc, 6$
	add hl, de
	adc c
6$:
	add hl, hl
	rla
	jr nc, 7$
	add hl, de
	adc c
7$:
	add hl, hl      ; 11
	rla             ;  4
	ret nc          ; 11 5
	add hl, de      ; 11
	adc c           ;  4
	ret             ; 10   AHL

;----------------------------------------------------------------------
; DEHL = HL * BC, fewer onebits preferably in BC
;    HL
;    BC
;  ----
;   AHL
;  AHL
;  ----
;  DEHL

mul_hl_bc::
	push hl
	ld a, c
	call mul_hl_a   ; lspart, keeps B

	pop de          ; original HL
	push af         ; save lspart
	push hl
	ld a, b         ; still original B
	call mul_de_a   ; mspart

	ld d, a         ; left 8 bits
	ld e, h
	ld h, l
	ld l, # 0       ; mspart aligned in DEHL

	pop bc          ; lspart
	pop af
	add hl, bc
	adc e
	ld e, a
	ret nc
	inc d
	ret

;----------------------------------------------------------------------
; HL * BC        DEHL
; DE * BC      DEHL
;              ------
;              BCDEHL

mul_dehl_bc::
	push de
	push bc

	call mul_hl_bc
	pop bc
	ex (sp), hl     ; lsword
	push de         ; part of middle word

	call mul_hl_bc  ; mspart

	ld b, d         ; msword in place
	ld c, e
	pop de
	add hl, de      ; merge middle words
	pop de          ; lsword
	ex de, hl       ; all in place
	ret nc
	inc bc          ; carry to msword
	ret

;----------------------------------------------------------------------
; DEHL / BC
; HL quotient
; DE remainder
; could be rearranged to use DJNZ XXX and split to 2 * 8

div_hl_bc::
	ld de, # 0

div_dehl_bc::
	ex de, hl         ; HLDE
	ld a, d           ; HLAE
	ld d, # 16        ; 16 times
2$:
	rl e              ; HLDE <<= 1, 1st carry flows out from result
	rla
	adc hl, hl
	jr c, 3$          ; 0x1???? - BC
	sbc hl, bc
	jr nc, 1$         ; ok
	add hl, bc        ; carry always reappears
1$:
	dec d
	jp nz, 2$         ; 16 times

	rl e              ; the final bit of quotient
	rla

	cpl
	ld d, a
	ld a, e
	cpl
	ld e, a

	ex de, hl
	ret

3$:
	and a             ; 0x1???? - BC
	sbc hl, bc
	and a             ; no carry
	jr 1$

;----------------------------------------------------------------------
; faking it, just scaling down arguments XXX

_divide_dwdwdw::
	ld  iy, # 2    ; skip ra
	add iy, sp
	jr  2$         ; start loop
1$:
	srl 7(iy)
	rr  6(iy)
	rr  5(iy)
	rr  4(iy)      ; divisor >>= 1

	srl 3(iy)
	rr  2(iy)
	rr  1(iy)
	rr  0(iy)      ; dividend >>= 1
2$:
	ld a, 7(iy)    ; divisor & 0xFFFF0000
	or a, 6(iy)
	jr nz, 1$      ; back, while divisor > 0xFFFF

_divide_dwdww::
	pop af
	pop de                ; dividend HLDE
	pop hl   
	pop bc                ; divisor BC
	push bc
	push hl
	push de
	push af

	push de          ; lsword
	call div_hl_bc
	ex (sp), hl      ; result msword <-> input lsword
	call div_dehl_bc
	push hl
	ex de, hl
	srl b
	rr c
	sbc hl, bc
	pop hl
	pop de
	ret c

	inc l               ; round up
	ret nz
	inc h
	ret nz
	inc de
	ret

;----------------------------------------------------------------------

ulmd_pow2:          ; A=C=128 or 64 (or 32 etc)
	dec a           ; mask
	and l           ; lsbits of input in A
1$:
	add hl, hl      ; align msbits to CDEH
	rl e
	rl d
	rl c
	jr nc, 1$       ; until the single onebit pops out

	ld 3(iy), c
	ld 2(iy), d
	ld 1(iy), e
	ld 0(iy), h

	ld l, a         ; remainder
	ret

_ulmoddiv::         ; byte ulmoddiv(dword *, byte);
	pop de
	pop iy          ; dword *inout
	pop bc          ; byte divisor
	push bc
	push iy
	push de

	ld d, 3(iy)
	ld e, 2(iy)
	ld h, 1(iy)
	ld l, 0(iy)

	ld a, c
	cp # 128        ; typical prescalers XXX not worth much
	jr z, ulmd_pow2
	cp # 64
	jr z, ulmd_pow2
	neg
	ld c, a         ; produce inverted carries

	call ulmd       ; churn
	rl e            ; mop last bit of msword
	rl d

	ld 3(iy), d
	ld 2(iy), e
	ld 1(iy), h
	ld 0(iy), l

	ld l, a         ; remainder
	ret

_umoddiv::          ; byte umoddiv(word *, byte);
	pop de
	pop iy          ; word *inout
	pop bc          ; byte divisor
	push bc
	push iy
	push de

	ld h, 1(iy)
	ld l, 0(iy)

	xor a
	sub c
	ld c, a         ; -divisor

	xor a           ; clear carry and remainder
	call umd

	ld 1(iy), h
	ld 0(iy), l

	ld l, a
	ret

;----------------------------------------------------------------------

ulmd:
	xor a            ; remainder, no carry

	inc d
	dec d
	jr nz, umd_3
	inc e
	dec e
	jr nz, umd_2
umd:                 ; XOR A by caller
	inc h
	dec h
	jr nz, umd_1
	jp umd_0

umd_3:
	ld b, d
	call umd8
	ld d, b
umd_2:
	ld b, e
	call umd8
	ld e, b
umd_1:
	ld b, h
	call umd8
	ld h, b
umd_0:
	ld b, l
	call umd8
	ld l, b

	adc hl, hl       ; mop last bit. if needed, RL DE by caller
	ret

umd8:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 1$         ; 12 or 7
	sub c            ;     + 4
1$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 2$         ; 12 or 7
	sub c            ;     + 4
2$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 3$         ; 12 or 7
	sub c            ;     + 4
3$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 4$         ; 12 or 7
	sub c            ;     + 4
4$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 5$         ; 12 or 7
	sub c            ;     + 4
5$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 6$         ; 12 or 7
	sub c            ;     + 4
6$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	jr c, 7$         ; 12 or 7
	sub c            ;     + 4
7$:
	rl b             ;  8
	rla              ;  4
	add c            ;  4
	ret c
	sub c
	ret

;----------------------------------------------------------------------

_format_hex::             ; byte *format_hex(byte *ep, word num);
	pop af
	pop de                ; end p
	push de
	push af

	ld hl, # 5            ; &num
	add hl, sp            ; msbyte
1$:
	xor a                 ; for once, give a chance to the funny opcodes
	rrd
	dec hl                ; lsbyte
	rrd                   ; A = num & 15; num >>= 4
	cp # 10               ; make ascii
	sbc # 0x69
	daa
	dec de
	ld (de), a            ; *--ep = hexdig[num & 15]

	ld a, (hl)
	inc hl                ; msbyte
	or (hl)
	jr nz, 1$             ; while (num)

	ex de, hl             ; return (ep)
	ret

_format_short::           ; byte *format_short(byte *ep, short num);
	pop af
	pop de                ; end p
	pop hl                ; num
	push hl
	push de
	push af

	bit 7, h
	jr z, fw              ; positive, nothing special

	call neg
	call fw               ; format_word(neg(num))

	dec hl                ; prepend sign
	ld (hl), # '-
	ret

_format_word::            ; byte *format_word(byte *ep, word num);
	pop af
	pop de                ; end p
	pop hl                ; num
	push hl
	push de
	push af
fw:
	ld c, # -10           ; HL /= 10
	jr fw_2               ; start from bottom
fw_1:
	xor a                 ; remainder
	call umd

	add # 0x30
	dec de
	ld (de), a            ; *--ep = remainder + '0'
fw_2:
	ld a, h
	or a
	jr nz, fw_1
	ld a, l
	cp # 10
	jr nc, fw_1           ; while (num > 9)

	add # 0x30
	dec de
	ld (de), a

	ex de, hl
	ret                   ; return (ep)

_format_dword::           ; byte *format_dword(byte *ep, dword num);
	pop af
	pop iy                ; end p
	pop hl                ; num
	pop de
	push de
	push hl
	push iy
	push af
fdw:
	ld c, # -10           ; DEHL /= 10
	jr 2$                 ; start from bottom
1$:
	call ulmd
	rl e                  ; mop last bit of msword
	rl d

	add # 0x30
	dec iy
	ld (iy), a            ; *--ep = remainder + '0'
2$:
	ld a, d
	or e
	jr nz, 1$             ; while (num > 65535)

	push iy
	pop de
	jp fw_2               ; HL holds num, continue in format_word

_format_long::
	pop af
	pop iy
	pop hl
	pop de
	push de
	push hl
	push iy
	push af

	bit 7, d
	jr z, fdw         ; positive, nothing special

	call lneg
	call fdw          ; format_dword(neg(dw))

	dec hl            ; prepend sign
	ld (hl), # '-
	ret

;----------------------------------------------------------------------

_crc_ccitt::       ; word crc_ccitt(byte *, word);
	pop hl
	pop de         ; byte *
	pop bc         ; len
	push bc
	push de
	push hl

crc_ccitt::
	ld b, c        ; length is 1...256 XXX

	ld a, # 0xFF
	ld c, a        ; iv FFFF in CA

	ld h, #>ccitt_tab_hi_lo
1$:
	ex de, hl
	xor (hl)       ; buf
	ex de, hl
	inc de

	ld l, a
	ld a, c
	ld c, (hl)     ; ccitt_tab_hi
	inc h
	xor (hl)       ; ccitt_tab_lo
	dec h
	djnz 1$

	cpl
	ld l, a
	ld a, c
	cpl
	ld h, a        ; ~crc in HL and AL, bufend in DE
	ret

;----------------------------------------------------------------------

_load_dword::
	pop af
	pop hl
	push hl
	push af

	di

	ld c, (hl)
	inc hl
	ld b, (hl)
	inc hl
	ld e, (hl)
	inc hl
	ld d, (hl)
	ld l, c
	ld h, b

	jp di_undo_trash_af

_load_word::
	pop af
	pop hl
	push hl
	push af

	di

	ld a, (hl)
	inc hl
	ld h, (hl)
	ld l, a

	jp di_undo_trash_af


_store_dword::
	pop af
	pop hl
	pop bc
	pop de
	push de
	push bc
	push hl
	push af

	di

	ld (hl), c
	inc hl
	ld (hl), b
	inc hl
	ld (hl), e
	inc hl
	ld (hl), d

	jp di_undo_trash_af

_store_word::      ; void store_word(word address, word value);
	pop af
	pop hl
	pop bc
	push bc
	push hl
	push af

	di

	ld (hl), c
	inc hl
	ld (hl), b

	jp di_undo_trash_af

;----------------------------------------------------------------------
	.IF 0 ; too much hassle

	pop af         ; ra
	pop de         ; address
	pop hl         ; value

	ld b, # 0xC9   ; RET
	ld c, d
	push bc        ; space of arguments is reused for the short code
	ld b, e
	ld c, # 0x22   ; LD (NN), HL
	push bc
	push af        ; ra back (Flags did not change)

	ld iy, # 2     ; point iy to the LD, skipping ra
	add iy, sp
	jp (iy)

_store_bit::       ; void store_bit(word address, byte bitno, bool nz);
	pop de         ; ra
	pop hl         ; address
	pop bc         ; c bitno, b value

	ld a, c
	and # ~7
	jr z, 1$       ; bit is within (hl)
	rrca
	rrca
	rrca
	add l
	ld l, a        ; reach for the correct byte
	jr nc, 1$
	inc h
1$:
	ld a, c
	and # 7
	rlca
	rlca
	rlca
	or # 0x86      ; RES: 0xCB, 0x86 | (bit << 3)
	inc b
	dec b
	jr z, 2$
	or # 0x40      ; SET: 0xCB, 0xC6 | (bit << 3)
2$:
	ld c, # 0xC9   ; RET
	push bc
	ld c, # 0xCB
	ld b, a
	push bc
	push de        ; ra back

	ld iy, # 2
	add iy, sp
	jp (iy)

	.ENDIF

;----------------------------------------------------------------------

	.IF 0

_setjmp::          ; int setjmp(jmp_buf b);
	pop de         ; ra
	pop iy         ; buf { sp pc ix }
	push iy        ; any junk would do XXX push junk instead
	push de

	ld hl, # 4     ; ra and arg not counted
	add hl, sp
	ld 0(iy), l    ; sp
	ld 1(iy), h
	ld 2(iy), e    ; pc
	ld 3(iy), d
	push ix
	pop hl
	ld 4(iy), l    ; ix
	ld 5(iy), h

	ld hl, # 0
	ret

_longjmp::         ; void longjmp(jmp_buf b, int retval);
	pop de         ; ra, discarded
	pop iy         ; buf { sp pc ix }
	pop de         ; retval

	ld b, 5(iy)
	ld c, 4(iy)    ; ix
	push bc
	ld b, 3(iy)
	ld c, 2(iy)    ; pc
	ld h, 1(iy)
	ld l, 0(iy)    ; sp
	pop ix
	ld sp, hl      ; after all data loaded, buf might vanish
	push iy        ; mimics setjmp args
	push bc        ; pc

	ex de, hl      ; retval to hl
	ret

	.ENDIF

;----------------------------------------------------------------------

_axaddr::          ; byte *axaddr(byte *dst, byte *s);
	pop bc
	pop hl         ; dest buf
	pop de         ; ascii string input
	push de
	push hl
	push bc

	ld b, # 6
3$:
	ld a, (de)
	inc de
	cp # 'a
	jr nc, 2$      ; a..z
	cp # '0
	jr nc, 1$      ; 0..9 or A..Z
	dec de         ; stop at - or \0
	xor a          ; put blanks
2$:
	xor # 0x20     ; lower to upper, \0 to blank
1$:
	add a          ; <<= 1
	ld (hl), a
	inc hl
	djnz 3$        ; 6 characters, then B zero

	ld a, (de)     ; "-SSID" ?
	or a
	jr z, 5$       ; no ssid, 0 in B ok
	cp # '-
	jr nz, 4$      ; accept "aa9bbb15"
	inc de
	ld a, (de)     ; or "aa9bbb-15"
4$:
	sub #'0
	cp # 10
	jr nc, 5$      ; not a digit, no ssid
	inc de
	ld b, a        ; 0 ... 9
	dec a
	jr nz, 5$      ; not 1, must be single digit
	ld a, (de)
	sub #'0
	cp # 6
	jr nc, 5$      ; single digit, after all
	inc de
	add # 10
	ld b, a        ; 10 ... 15
5$:
	ld a, b
	or # 0x30      ; reserved ssid bits
	add a          ; <<= 1
	ld (hl), a
	inc hl         ; return end of dst buffer
	ret            ; end of input in de

;----------------------------------------------------------------------
; while (--arg)
;   ;

_delay_usec::          ; void delay_usec(word usec);
	pop hl
	pop de             ; arg
	push de
	push hl

	srl d              ; half
	rr e
	ld a, (_logic)
	or a
	jr z, 1$           ; P8E, loop takes 2.48 usec
	xor a              ; P8N, loop takes 4.96 usec
	srl d              ; quarter
	rr e
1$:
	dec de
	or d
	jp p, 1$           ; 20 or 23 T total

	ret                ; additional 15 or 30 usec from call/setup/return

;----------------------------------------------------------------------

_timex::               ; dword timex(); seconds
1$:
	ld hl, (timex+0)
	ld de, (timex+2)
	ld a,  (timex+0)
	cp l
	ret z
	jr 1$              ; changed, try again

_time::                ; word time(); seconds
	ld hl, (time)
	ret

_timef::               ; word timef(); 1/256 seconds
	ld hl, (timef)
	ret

;----------------------------------------------------------------------

_timef_mark::          ; word timef_mark(word interval);
	pop hl
	pop de
	push de
	push hl

	ld hl, (timef)
	add hl, de
	set 0, l           ; always nonzero result
	ret

_time_mark::           ; word time_mark(word interval);
	pop hl
	pop de
	push de
	push hl

	ld hl, (time)
	add hl, de
	set 0, l           ; always nonzero result
	ret

;----------------------------------------------------------------------

_timef_passed::        ; bool timef_passed(word mark);
	pop hl
	pop de
	push de
	push hl

	xor a              ; clear carry, and FALSE
	ld hl, (timef)
	sbc hl, de         ; timef - mark
	ld l, a
	ret m              ; FALSE
	inc l              ; TRUE
	ret

_time_passed::         ; bool time_passed(word mark);
	pop hl
	pop de
	push de
	push hl

	xor a
	ld hl, (time)
	sbc hl, de         ; time - mark
	ld l, a
	ret m
	inc l
	ret

;----------------------------------------------------------------------
; checksum, split and store field pointers to _nmea[]
; return 0 (int) if ok
; extra fields ignored

_nmea_split::         ; int nmea_split(char *s);
	pop de
	pop hl            ; asciz
	push hl
	push de
nmea_split::

	ld iy, # _nmea    ; field pointer array
	ld bc, # 0x0F00   ; b nfld  c chk
	ld de, # 0x2C2A   ; d comma e star
	jp 3$             ; start the loop
8$:
	inc b             ; back to 0
1$:
	inc hl
2$:
	xor c
	ld c, a
3$:
	ld a, (hl)
	cp e
	jr z, 4$          ; found the star
	or a
	ret z             ; \0. no star. failure
	cp d
	jp nz, 1$         ; not comma either, next character

	ld (hl), # 0      ; cut at comma

	dec b
	jp m, 8$          ; array full, just run the checksum to end

	inc hl            ; next field is here
	ld 0(iy), l       ; store it
	ld 1(iy), h
	inc iy
	inc iy
	jp 2$             ; continue scan

4$:
	ld (hl), # 0      ; cut at star

	call 6$           ; does not return if bad hex
	rlca
	rlca
	rlca
	rlca
	ld e, a           ; msbits of checksum

	call 6$           ; does not return if bad hex
	or e              ; complete checksum
	xor c
	ret nz            ; checksum wrong

	ld hl, # 0        ; success return and NULL
5$:
	dec b             ; array full ?
	ret m             ; success
	ld 0(iy), l       ; append NULL
	ld 1(iy), h
	inc iy
	inc iy
	jp 5$             ; until array full

6$:
	inc hl            ; checksum character
	ld a, (hl)
	sub # '0
	jr c, 7$          ; bad hex
	cp # 10
	ret c             ; 0...9

	sub # 'A-'0
	cp # 6
	jr nc, 7$         ; bad hex
	add # 10          ; A...F
	ret
7$:
	pop af            ; abort one level
	ret               ; hl still nz, failure

;----------------------------------------------------------------------
	.IF 0
;----------------------------------------------------------------------
; setup thinks all numbers are base 10,
; but this one is base 8, max 0777.
; HL in/out

fix_octal:
	ld a, l
	sub # 778
	ld a, h
	sbc # 778 >> 8
	jr c, 1$

	ld hl, # 0x01FF
	ret                 ; overflow. return max
1$:
	ld de, # 0xFFFF     ; -1 in D and E
	ld bc, # -100
2$:
	inc d
	add hl, bc
	jp c, 2$
	sbc hl, bc

	ld a, l             ; max 99 now
	ld c, # 10
3$:
	inc e
	sub c
	jp nc, 3$
	add c               ; bit triplets in D E and A

	rrca                ; position lsbits to 0xE0
	rrca
	rrca
	or e                ; lsbits 0 0 middlebits
	rrca
	rrca
	rrca                ; middlebits lsbits 0 0
	srl d
	rra
	srl d
	rra
	ld h, d             ; zeroextended msbit
	ld l, a             ; 8 lsbits
	ret

;----------------------------------------------------------------------

_dcs_frame::            ; dword dcs_frame(int code);
	pop de
	pop hl
	push hl
	push de

	push hl             ; code +- 0...777
	call _abs           ; argument not popped, will be used later
	call fix_octal      ; in HL. code was OCTAL, not a decimal number

	; binary rightjust +code in HL

	; frame looks like P11...1 100 C9...1
	; and C1 is sent first
	;
	; get the full 23 bits to E:H:L
	;
	; E   0  P11 P10 P9 - P8  P7  P6  P5
	; H   P4 P3  P2  P1 - 1   0   0   C9
	; L   C8 C7  C6  C5 - C4  C3  C2  C1

	ld de, # 0x8000   ; d collects P1...4 from left and e P11..5 from right

	; mucho gracias MMI Technical Resource website. good shit.

	; P1 = C1 + C2 + C3 + C4 + C5 + C8

	ld a, l           ; 8 lsbits
	and # 0x9F        ; AND clears carry and tests parity
	jp pe, .+4        ; if even parity, skip SCF
	scf               ; CCF would also do
	rr d              ; insert P1

	; P2 = NOT ( C2 + C3 + C4 + C5 + C6 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0x9F
	jp po, .+4
	scf
	rr d

 	; P3 = C1 + C2 + C6 + C7 + C8

	ld a, l
	and # 0xE3
	jp pe, .+4
	scf
	rr d

 	; P4 = NOT ( C2 + C3 + C7 + C8 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0xE3
	jp po, .+4
	scf
	rr d

	; P11 = NOT ( C1 + C2 + C3 + C4 + C7 )

	ld a, l
	and # 0x4F
	jp po, .+4
	scf
	rl e

	; P10 = NOT ( C4 + C6 + C7 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0xB4
	jp po, .+4
	scf
	rl e

 	; P9 = C3 + C5 + C6 + C8 + C9

	ld a, h
	rra
	ld a, l
	rra
	and # 0xDA
	jp pe, .+4
	scf
	rl e

 	; P8 = C2 + C4 + C5 + C7 + C8 + C9

	ld a, h
	rra
	ld a, l
	rra
	and # 0xED
	jp pe, .+4
	scf
	rl e

 	; P7 = C1 + C3 + C4 + C6 + C7 + C8 + C9

	ld a, l
	and # 0xED     ; C8...C1
	jp pe, .+4
	scf            ; C1 +...+ C8

	bit 0, h
	jr z, .+3
	ccf            ; + C9
	rl e

 	; P6 = NOT ( C1 + C4 + C5 + C6 + C8 )

	ld a, l
	and # 0xB9
	jp po, .+4
	scf
	rl e

 	; P5 = NOT ( C1 + C2 + C5 + C9 )

	ld a, h
	rra
	ld a, l
	rla          ; drop C8, A contains C7 ... C1 C9
	and # 0x27
	jp po, .+4
	scf
	rl e

	; whew.

	ld a, h        ; just C9
	or d           ; add 1 0 0 flag and parity lsbits
	ld h, a

	pop af         ; was argument for abs() long time ago
	rla
	jr nc, 1$      ; +code
	ld a, l
	cpl
	ld l, a
	ld a, h
	cpl
	ld h, a
	ld a, e
	cpl
	ld e, a        ; -code, bits inverted
1$:
	res 7, e
	ld d, # 0
	ret            ; in 23 lsbits of DEHL

;----------------------------------------------------------------------
	.ENDIF
;----------------------------------------------------------------------

_make_sine::        ; void make_sine(byte *tab, byte gain, byte center)
	pop hl
	pop iy          ; iy tab
	pop de          ; d center and e gain
	push de
	push iy
	push hl
make_sine::         ; IY, D and E

	ld a, d         ; center
	or a            ; less than 128 ?
	jp p, 1$        ; distance to 0,
	cpl             ;  to 255, is shorter
1$:
	add a, a        ; twice, max available amplitude
	cp e            ; gain
	jr nc, 2$
	ld e, a         ; override gain
2$:
	push ix               ; save ix

	ld bc, # 0xC0
	add iy, bc            ; 4th quadrant
	push iy
	pop ix                ; equals end of 3rd q

	ld bc, # _sine + 0x3F ; last of 1st q
4$:
	dec ix                ; 1st and 3rd step backwards

	ld a, (bc)            ; sine[]
	ld h, a
	ld a, d               ; save center

	call mul_h_e          ; HL=H*E, H = sine * gain / 256, keeps A BC E

	ld d, a               ; center
	sub h                 ; center - value
	ld (ix), a            ; 3rd q
	ld (iy), a            ; 4th q

	ld a, d               ; center
	add h                 ; center + value
	ld -128(ix), a        ; 1st q
	ld -128(iy), a        ; 2nd q

	inc iy                ; 2nd and 4th step forward
	dec c
	jp p, 4$              ; page-aligned sine[3F ... 0]

	pop ix
	ret

;----------------------------------------------------------------------

_legal_tx::             ; byte legal_tx(dword kHz);
	pop hl
	pop bc
	pop de              ; kHz in DEBC
	push de
	push bc
	push hl

	bit 7, d            ; the loop below cannot handle ~0 kHz,
	jr nz, illegal      ; cheat, and check the msbit.

	ld hl, # _cfg + tx_band_start  ; followed by tx_band_end
	call limit
	jr c, illegal
	call limit
	jr nc, illegal

	ld hl, # _tx_bands  ; { start, end, ..., end, ~0 } in ascending order
1$:
	call limit
	jr c, illegal
	call limit
	jr nc, 1$           ; at or above end, try next

	ld l, # 1           ; legal
	ret

illegal:
	ld l, # 0
	ret

limit:
	ld a, c             ; unknown - limit
	sub (hl)
	inc hl              ; walks the list
	ld a, b
	sbc (hl)
	inc hl
	ld a, e
	sbc (hl)
	inc hl
	ld a, d
	sbc (hl)
	inc hl
	ret                 ; c if below limit

;----------------------------------------------------------------------
;  TMR_LPF is MF6-100 clock
;
;  FFSK input goes via the low-pass filter,
;  so LPF clock must run continuously.
;  Funny though... FFSK receive DID work (marginally?)
;  even with LPF clock stopped. XXX ???
;
;  Update:
;  FFSKI is taken from INPUT of LP filter XXX ?!?
;
;                                  LPFCLK
;                                     |
;                      MF6........................
;                      .              |          .
;  MIC -----------+--------AMP---+---LPF---AMP-------- MOD
;  CCIR(=pwm) ----+    .         |               .
;  FFSKO ---------+    ...........................
;                                |
;                                +--- FFSKI
;
;  Above only for L8TM (L8M?)
;  On P8x (A8N really) LPF and FFSKI have no connection.
;----------------------------------------------------------------------

LPF_ccir::                   ; HL Hz
	ld a, (_cfg + ccir_gain) ; play with LPF slope ?
	or a
	jr z, LPF                ; 0 means no tricks

	call mul_hl_a            ; AHL = A * HL
	ld l, h
	ld h, a                  ; HL = Hz * ccir_gain / 256
	jr LPF

_LPF::                      ; void LPF(word hz);
	pop de
	pop hl                  ; hz
	push hl
	push de
LPF::                       ; hz in HL

	ld a, h
	or l
	jr z, LPF_off           ; zero hz, stop timer

	ld c, l                 ; (4032000 / 100 + Hz / 2) / Hz
	ld b, h
	srl h
	rr l
	ld de, # TMR_CLK / 100
	add hl, de
	call div_hl_bc

	ld a, # TMR_SEL_LPF | TMR_BOTH | TMR_SQWAVE

	di

	out (TMR_CTRL), a
	ld a, l
	out (TMR_LPF), a
	ld a, h
	out (TMR_LPF), a

	jp di_undo_trash_af

LPF_off:
	ld a, # TMR_SEL_LPF | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a
	ret                     ; output constant 0

; sidetone and ccir tx

_timer_mt::                 ; void timer_mt(word hz)
	pop de
	pop hl                  ; hz
	push hl
	push de
timer_mt::

	ld a, h
	or l
	jr z, _timer_mt_off

	ld c, l
	ld b, h
	ld de, # TMR_CLK / 65536
	ld hl, # TMR_CLK % 65536
	call div_dehl_bc          ; no rounding

	ld a, # TMR_SEL_MT | TMR_BOTH | TMR_SQWAVE

	di

	out (TMR_CTRL), a
	ld a, l
	out (TMR_MT), a
	ld a, h
	out (TMR_MT), a

	jp di_undo_trash_af

_timer_mt_off::
	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a
	ret                     ; output constant 0

;----------------------------------------------------------------------

_dtmf::                 ; ? dtmf(?)
	ld hl, (_cu_dtmf)
	jp (hl)

;----------------------------------------------------------------------

_tx_led_on::
	ld hl, # _leds
	set LED_ROAM, (hl)
	ret

_tx_led_off::
	ld hl, # _leds
	res LED_ROAM, (hl)
	ret

_call_led_on::
	ld hl, # _leds
	set LED_CALL, (hl)
	ret

_call_led_off::
	ld hl, # _leds
	res LED_CALL, (hl)
	ret

_lights_on::
	ld hl, (_cfg + lights)
	ld a, h
	or l
	ret z                     ; timeout 0, no lights

	ld (light_timo), hl       ; XXX if 1, tick_1Hz might sneak in

	ld hl, # _leds
	set LED_LCDLIGHT, (hl)
	set LED_KEYLIGHT, (hl)
	ret

;----------------------------------------------------------------------

_open_audio::
	ld a, (_cfg + volume)      ; zero volume ?
	or a
	ret z                      ; is silent

	ld de, # O0_AUDIOC * 0x101 ; AUDIOC on
	jp out_0_change

_close_audio::
	ld de, # O0_AUDIOC ;  AUDIOC off
	jp out_0_change


_close_squelch::
	ld a, (_audio_flags)
	or a
	call z, _close_audio       ; close audio unless squelch flags active

	call rewind_squelch
	xor a
	ld (_squelch_open), a

	ld hl, # _leds
	res LED_SERV, (hl)

	ld de, # PB_EXIN1 * 0x101  ; EXIN1 input mode, pullup to +5v
	jp pio_b_change_ddr

_open_squelch::
	ld a, (_audio_flags)
	or a
	call z, _open_audio       ; open audio unless squelch flags active

	call rewind_squelch
	ld a, # 1
	ld (_squelch_open), a

	ld hl, # _leds
	set LED_SERV, (hl)

	ld de, # PB_EXIN1         ; EXIN1 output 0
	jp pio_b_change_ddr

;----------------------------------------------------------------------
; 0 silent service
; 1 ear only
; 8 speaker levels

_volume::
	pop hl
	pop de      ; e -1, 0, +1
	push de
	push hl

	ld hl, # _cfg + volume  ; old volume should be 0 ... 9
	ld a, (hl)
	add e
	jp p, 0$
	xor a                   ; undo 0 -= 1
0$:
	cp # 10
	jr c, 9$
	ld a, # 9               ; max volume is 9
9$:
	ld (hl), a

	ld d, # O0_INH          ; no speaker, unless volume >= 2
	sub # 2
	jr c, 2$
	ld d, a                 ; speaker is audible, no INH
2$:
	ld hl, (_cu_volume)     ; CU58AF ?
	ld a, h
	or l
	jr z, 3$                ; CU53

	rst JPHL                ; cu_volume returns appropriate O0_ bits
	ld d, l
3$:
	ld e, # O0_INH | O0_VOLUME ; change speaker volume bits
	call out_0_change

	rst DISABLE
	call 1$                 ; protected opening/closing of audio, maybe.
	jp _restore             ; done

1$:
	ld a, (_cfg + volume)
	or a
	jp z, _close_audio      ; zero volume, always muted

	ld a, (_audio_flags)    ; else, squelch open or monitor on ?
	and # 3
	ld hl, # _squelch_open
	or (hl)
	ret z                   ; no, leave audio closed
	jp _open_audio

;----------------------------------------------------------------------

_fx_check::
	pop hl
	pop de        ; byte *p
	pop bc        ; word datalen
	push bc
	push de
	push hl

	call crc_ccitt  ; ~crc in HL and AL, bufend in DE

	ld a, (de)      ; MSByte
	sub h
	jr nz, 1$
	inc de
	ld a, (de)      ; LSByte
	sub l
1$:
	ld l, a         ; 0 if ok
	ret

_fx_send::
	call _buf_alloc
	ld a, h
	or l
	jp z, _crash

	pop af            ; arguments;
	pop de            ; byte *data
	pop bc            ; word datalen
	push bc
	push de
	push af

	ld a, # 3 + 2     ; length; 3 for synch and 2 for crc
	add c

	push hl           ; buf
	push bc           ; datalen

	inc hl            ; 'next' is zeroed by buf_alloc
	inc hl

	ld (hl), a        ; 'len'
	inc hl
	ld (hl), # 0      ; 'type' FSK
	inc hl

	ld (hl), # 0xAA   ; 8 idlebits between packets
	inc hl
	ld (hl), # 0xC4   ; SYNC
	inc hl
	ld (hl), # 0xD7
	inc hl

	push hl           ; data start in buf
	ex de, hl
	ldir

	pop de            ; data
	pop bc            ; datalen
	call crc_ccitt    ; ~crc in HL, bufend in DE
	ex de, hl
	ld (hl), d        ; MSByte
	inc hl
	ld (hl), e        ; packet done.

	pop de            ; buf

	rst DISABLE

	ld hl, (_fx_tx_tail)     ; tail = *tail = buf
	ld (_fx_tx_tail), de
	ld (hl), e
	inc hl
	ld (hl), d

	.IF P8x                  ; start modem, if needed

	ld a, (_fx429_cmd)       ; active ?
	bit 0, a                 ; FX429_TXENB
	jr nz, 1$                ; yes
	or # 1
	ld (_fx429_cmd), a
	out (FX429_CSR), a

	ld a, # 0xAA             ; 16 additional idlebits
	out (FX429_DATA), a
	out (FX429_DATA), a      ; fills holding register

	.ENDIF

	.IF L8x

	ld a, (_sio_a_wr5)       ; active ?
	bit 1, a                 ; SA_TXE aka WR5_RTS
	jr nz, 1$                ; yes
	or # 2
	ld (_sio_a_wr5), a

	ld bc, # 0x500 + SIO_A_CSR
	in e, (c)                ; soita paranoid
	out (c), b               ; WR5
	out (c), a

	ld a, # 0x55             ; reversed 0xAA
	out (SIO_A_DATA), a      ; there is no holding register in SIO/0

	; TX Underrun not reset, crc is manual (yes, the leg hurts)

	.ENDIF
1$:
	jp _restore

;----------------------------------------------------------------------
; cBels = centibel(rssi)
; displayed = (rssi - cfg.dB_zero) * cfg.dB_scale / 32
; RSSI is directly connected to ADC
; max scale 255 gives max displayed value about 2000 (or -2000)
;
; SA604A "typical"
;  -118 dBm =  160mV
;   -68 dBm = 2650mV
;   -18 dBm = 4850mV
; from the graph
;  -120 dBm 0v
;   -30 dBm 5v
;  0x00  0 dB
;  0xFF 90 dB

_centibel::
	pop hl
	pop de             ; rssi in E
	push de
	push hl

	ld hl, (_cfg + dB_zero)        ; zero in L, scale in H
	ld a, e
	sub l                          ; rssi - zero
	jr nc, 1$
	neg                            ; below zero, negate
	call 1$                        ; scale
	jp neg                         ; return negative short
1$:
	ld e, a

	call mul_h_e      ; scale * (rssi - zero)

	xor a             ; divide by 32 ...
	add hl, hl
	rla
	add hl, hl
	rla
	add hl, hl
	rla               ; *= 8
	ld l, h
	ld h, a           ; /= 256, gives divide by 32
	ret               ; positive short in HL

; XXX dB_zero as a word, would allow dBm display.
;----------------------------------------------------------------------
